#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python wrapper for Android uiautomator tool."""

import sys
import os
import subprocess
import time
import itertools
import json
import hashlib
import socket
import re
import collections
import xml.dom.minidom

DEVICE_PORT = int(os.environ.get('UIAUTOMATOR_DEVICE_PORT', '9008'))
LOCAL_PORT = int(os.environ.get('UIAUTOMATOR_LOCAL_PORT', '9008'))

if 'localhost' not in os.environ.get('no_proxy', ''):
    os.environ['no_proxy'] = "localhost,%s" % os.environ.get('no_proxy', '')

try:
    import urllib2
except ImportError:
    import urllib.request as urllib2
try:
    from httplib import HTTPException
except:
    from http.client import HTTPException
try:
    if os.name == 'nt':
        import urllib3
except:  # to fix python setup error on Windows.
    pass

__author__ = "Xiaocong He"
__all__ = ["device", "Device", "rect", "point", "Selector", "JsonRPCError"]


def U(x):
    if sys.version_info.major == 2:
        return x.decode('utf-8') if type(x) is str else x
    elif sys.version_info.major == 3:
        return x


def param_to_property(*props, **kwprops):
    if props and kwprops:
        raise SyntaxError("Can not set both props and kwprops at the same time.")

    class Wrapper(object):

        def __init__(self, func):
            self.func = func
            self.kwargs, self.args = {}, []

        def __getattr__(self, attr):
            if kwprops:
                for prop_name, prop_values in kwprops.items():
                    if attr in prop_values and prop_name not in self.kwargs:
                        self.kwargs[prop_name] = attr
                        return self
            elif attr in props:
                self.args.append(attr)
                return self
            raise AttributeError("%s parameter is duplicated or not allowed!" % attr)

        def __call__(self, *args, **kwargs):
            if kwprops:
                kwargs.update(self.kwargs)
                self.kwargs = {}
                return self.func(*args, **kwargs)
            else:
                new_args, self.args = self.args + list(args), []
                return self.func(*new_args, **kwargs)
    return Wrapper


class JsonRPCError(Exception):

    def __init__(self, code, message):
        self.code = int(code)
        self.message = message

    def __str__(self):
        return "JsonRPC Error code: %d, Message: %s" % (self.code, self.message)


class JsonRPCMethod(object):

    if os.name == 'nt':
        try:
            pool = urllib3.PoolManager()
        except:
            pass

    def __init__(self, url, method, timeout=30):
        self.url, self.method, self.timeout = url, method, timeout

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise SyntaxError("Could not accept both *args and **kwargs as JSONRPC parameters.")
        data = {"jsonrpc": "2.0", "method": self.method, "id": self.id()}
        if args:
            data["params"] = args
        elif kwargs:
            data["params"] = kwargs
        jsonresult = {"result": ""}
        if os.name == "nt":
            res = self.pool.urlopen("POST",
                                    self.url,
                                    headers={"Content-Type": "application/json"},
                                    body=json.dumps(data).encode("utf-8"),
                                    timeout=self.timeout)
            jsonresult = json.loads(res.data.decode("utf-8"))
        else:
            result = None
            try:
                req = urllib2.Request(self.url,
                                      json.dumps(data).encode("utf-8"),
                                      {"Content-type": "application/json"})
                result = urllib2.urlopen(req, timeout=self.timeout)
                jsonresult = json.loads(result.read().decode("utf-8"))
            finally:
                if result is not None:
                    result.close()
        if "error" in jsonresult and jsonresult["error"]:
            raise JsonRPCError(
                jsonresult["error"]["code"],
                "%s: %s" % (jsonresult["error"]["data"]["exceptionTypeName"], jsonresult["error"]["message"])
            )
        return jsonresult["result"]

    def id(self):
        m = hashlib.md5()
        m.update(("%s at %f" % (self.method, time.time())).encode("utf-8"))
        return m.hexdigest()


class JsonRPCClient(object):

    def __init__(self, url, timeout=30, method_class=JsonRPCMethod):
        self.url = url
        self.timeout = timeout
        self.method_class = method_class

    def __getattr__(self, method):
        return self.method_class(self.url, method, timeout=self.timeout)


class Selector(dict):

    """The class is to build parameters for UiSelector passed to Android device.
    """
    __fields = {
        "text": (0x01, None),  # MASK_TEXT,
        "textContains": (0x02, None),  # MASK_TEXTCONTAINS,
        "textMatches": (0x04, None),  # MASK_TEXTMATCHES,
        "textStartsWith": (0x08, None),  # MASK_TEXTSTARTSWITH,
        "className": (0x10, None),  # MASK_CLASSNAME
        "classNameMatches": (0x20, None),  # MASK_CLASSNAMEMATCHES
        "description": (0x40, None),  # MASK_DESCRIPTION
        "descriptionContains": (0x80, None),  # MASK_DESCRIPTIONCONTAINS
        "descriptionMatches": (0x0100, None),  # MASK_DESCRIPTIONMATCHES
        "descriptionStartsWith": (0x0200, None),  # MASK_DESCRIPTIONSTARTSWITH
        "checkable": (0x0400, False),  # MASK_CHECKABLE
        "checked": (0x0800, False),  # MASK_CHECKED
        "clickable": (0x1000, False),  # MASK_CLICKABLE
        "longClickable": (0x2000, False),  # MASK_LONGCLICKABLE,
        "scrollable": (0x4000, False),  # MASK_SCROLLABLE,
        "enabled": (0x8000, False),  # MASK_ENABLED,
        "focusable": (0x010000, False),  # MASK_FOCUSABLE,
        "focused": (0x020000, False),  # MASK_FOCUSED,
        "selected": (0x040000, False),  # MASK_SELECTED,
        "packageName": (0x080000, None),  # MASK_PACKAGENAME,
        "packageNameMatches": (0x100000, None),  # MASK_PACKAGENAMEMATCHES,
        "resourceId": (0x200000, None),  # MASK_RESOURCEID,
        "resourceIdMatches": (0x400000, None),  # MASK_RESOURCEIDMATCHES,
        "index": (0x800000, 0),  # MASK_INDEX,
        "instance": (0x01000000, 0)  # MASK_INSTANCE,
    }
    __mask, __childOrSibling, __childOrSiblingSelector = "mask", "childOrSibling", "childOrSiblingSelector"

    def __init__(self, **kwargs):
        super(Selector, self).__setitem__(self.__mask, 0)
        super(Selector, self).__setitem__(self.__childOrSibling, [])
        super(Selector, self).__setitem__(self.__childOrSiblingSelector, [])
        for k in kwargs:
            self[k] = kwargs[k]

    def __setitem__(self, k, v):
        if k in self.__fields:
            super(Selector, self).__setitem__(U(k), U(v))
            super(Selector, self).__setitem__(self.__mask, self[self.__mask] | self.__fields[k][0])
        else:
            raise ReferenceError("%s is not allowed." % k)

    def __delitem__(self, k):
        if k in self.__fields:
            super(Selector, self).__delitem__(k)
            super(Selector, self).__setitem__(self.__mask, self[self.__mask] & ~self.__fields[k][0])

    def clone(self):
        kwargs = dict((k, self[k]) for k in self
                      if k not in [self.__mask, self.__childOrSibling, self.__childOrSiblingSelector])
        selector = Selector(**kwargs)
        for v in self[self.__childOrSibling]:
            selector[self.__childOrSibling].append(v)
        for s in self[self.__childOrSiblingSelector]:
            selector[self.__childOrSiblingSelector].append(s.clone())
        return selector

    def child(self, **kwargs):
        self[self.__childOrSibling].append("child")
        self[self.__childOrSiblingSelector].append(Selector(**kwargs))
        return self

    def sibling(self, **kwargs):
        self[self.__childOrSibling].append("sibling")
        self[self.__childOrSiblingSelector].append(Selector(**kwargs))
        return self

    child_selector, from_parent = child, sibling


def rect(top=0, left=0, bottom=100, right=100):
    return {"top": top, "left": left, "bottom": bottom, "right": right}


def intersect(rect1, rect2):
    top = rect1["top"] if rect1["top"] > rect2["top"] else rect2["top"]
    bottom = rect1["bottom"] if rect1["bottom"] < rect2["bottom"] else rect2["bottom"]
    left = rect1["left"] if rect1["left"] > rect2["left"] else rect2["left"]
    right = rect1["right"] if rect1["right"] < rect2["right"] else rect2["right"]
    return left, top, right, bottom


def point(x=0, y=0):
    return {"x": x, "y": y}


class Adb(object):

    def __init__(self, serial=None, adb_server_host=None, adb_server_port=None):
        self.__adb_cmd = None
        self.default_serial = serial if serial else os.environ.get("ANDROID_SERIAL", None)
        self.adb_server_host = str(adb_server_host if adb_server_host else 'localhost')
        self.adb_server_port = str(adb_server_port if adb_server_port else '5037')
        self.adbHostPortOptions = []
        if self.adb_server_host not in ['localhost', '127.0.0.1']:
            self.adbHostPortOptions += ["-H", self.adb_server_host]
        if self.adb_server_port != '5037':
            self.adbHostPortOptions += ["-P", self.adb_server_port]

    def adb(self):
        if self.__adb_cmd is None:
            if "ANDROID_HOME" in os.environ:
                filename = "adb.exe" if os.name == 'nt' else "adb"
                adb_cmd = os.path.join(os.environ["ANDROID_HOME"], "platform-tools", filename)
                if not os.path.exists(adb_cmd):
                    raise EnvironmentError(
                        "Adb not found in $ANDROID_HOME path: %s." % os.environ["ANDROID_HOME"])
            else:
                import distutils
                if "spawn" not in dir(distutils):
                    import distutils.spawn
                adb_cmd = distutils.spawn.find_executable("adb")
                if adb_cmd:
                    adb_cmd = os.path.realpath(adb_cmd)
                else:
                    raise EnvironmentError("$ANDROID_HOME environment not set.")
            self.__adb_cmd = adb_cmd
        return self.__adb_cmd

    def cmd(self, *args, **kwargs):
        '''adb command, add -s serial by default. return the subprocess.Popen object.'''
        serial = self.device_serial()
        if serial:
            if " " in serial:  # TODO how to include special chars on command line
                serial = "'%s'" % serial
            return self.raw_cmd(*["-s", serial] + list(args))
        else:
            return self.raw_cmd(*args)

    def raw_cmd(self, *args):
        '''adb command. return the subprocess.Popen object.'''
        cmd_line = [self.adb()] + self.adbHostPortOptions + list(args)
        if os.name != "nt":
            cmd_line = [" ".join(cmd_line)]
        return subprocess.Popen(cmd_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def device_serial(self):
        if not self.default_serial:
            devices = self.devices()
            if devices:
                if len(devices) is 1:
                    self.default_serial = list(devices.keys())[0]
                else:
                    raise EnvironmentError("Multiple devices attached but default android serial not set.")
            else:
                raise EnvironmentError("Device not attached.")
        return self.default_serial

    def devices(self):
        '''get a dict of attached devices. key is the device serial, value is device name.'''
        out = self.raw_cmd("devices").communicate()[0].decode("utf-8")
        match = "List of devices attached"
        index = out.find(match)
        if index < 0:
            raise EnvironmentError("adb is not working.")
        return dict([s.split("\t") for s in out[index + len(match):].strip().splitlines() if s.strip()])

    def forward(self, local_port, device_port):
        '''adb port forward. return 0 if success, else non-zero.'''
        return self.cmd("forward", "tcp:%d" % local_port, "tcp:%d" % device_port).wait()

    def forward_list(self):
        '''adb forward --list'''
        version = self.version()
        if int(version[1]) <= 1 and int(version[2]) <= 0 and int(version[3]) < 31:
            raise EnvironmentError("Low adb version.")
        lines = self.raw_cmd("forward", "--list").communicate()[0].decode("utf-8").strip().splitlines()
        return [line.strip().split() for line in lines]

    def version(self):
        '''adb version'''
        match = re.search(r"(\d+)\.(\d+)\.(\d+)", self.raw_cmd("version").communicate()[0].decode("utf-8"))
        return [match.group(i) for i in range(4)]


_init_local_port = LOCAL_PORT - 1


def next_local_port(adbHost=None):
    def is_port_listening(port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((str(adbHost) if adbHost else '127.0.0.1', port))
        s.close()
        return result == 0
    global _init_local_port
    _init_local_port = _init_local_port + 1 if _init_local_port < 32764 else LOCAL_PORT
    while is_port_listening(_init_local_port):
        _init_local_port += 1
    return _init_local_port


class NotFoundHandler(object):

    '''
    Handler for UI Object Not Found exception.
    It's a replacement of UiAutomator watcher on device side.
    '''

    def __init__(self):
        self.__handlers = collections.defaultdict(lambda: {'on': True, 'handlers': []})

    def __get__(self, instance, type):
        return self.__handlers[instance.adb.device_serial()]


class AutomatorServer(object):

    """start and quit rpc server on device.
    """
    __jar_files = {
        "bundle.jar": "libs/bundle.jar",
        "uiautomator-stub.jar": "libs/uiautomator-stub.jar"
    }

    __apk_files = ["libs/app-uiautomator.apk", "libs/app-uiautomator-test.apk"]

    __sdk = 0

    handlers = NotFoundHandler()  # handler UI Not Found exception

    def __init__(self, serial=None, local_port=None, device_port=None, adb_server_host=None, adb_server_port=None):
        self.uiautomator_process = None
        self.adb = Adb(serial=serial, adb_server_host=adb_server_host, adb_server_port=adb_server_port)
        self.device_port = int(device_port) if device_port else DEVICE_PORT
        if local_port:
            self.local_port = local_port
        else:
            try:  # first we will try to use the local port already adb forwarded
                for s, lp, rp in self.adb.forward_list():
                    if s == self.adb.device_serial() and rp == 'tcp:%d' % self.device_port:
                        self.local_port = int(lp[4:])
                        break
                else:
                    self.local_port = next_local_port(adb_server_host)
            except:
                self.local_port = next_local_port(adb_server_host)

    def push(self):
        base_dir = os.path.dirname(__file__)
        for jar, url in self.__jar_files.items():
            filename = os.path.join(base_dir, url)
            self.adb.cmd("push", filename, "/data/local/tmp/").wait()
        return list(self.__jar_files.keys())

    def install(self):
        base_dir = os.path.dirname(__file__)
        for apk in self.__apk_files:
            self.adb.cmd("install", "-r -t", os.path.join(base_dir, apk)).wait()

    @property
    def jsonrpc(self):
        return self.jsonrpc_wrap(timeout=int(os.environ.get("jsonrpc_timeout", 90)))

    def jsonrpc_wrap(self, timeout):
        server = self
        ERROR_CODE_BASE = -32000

        def _JsonRPCMethod(url, method, timeout, restart=True):
            _method_obj = JsonRPCMethod(url, method, timeout)

            def wrapper(*args, **kwargs):
                URLError = urllib3.exceptions.HTTPError if os.name == "nt" else urllib2.URLError
                try:
                    return _method_obj(*args, **kwargs)
                except (URLError, socket.error, HTTPException) as e:
                    if restart:
                        server.stop()
                        server.start(timeout=30)
                        return _JsonRPCMethod(url, method, timeout, False)(*args, **kwargs)
                    else:
                        raise
                except JsonRPCError as e:
                    if e.code >= ERROR_CODE_BASE - 1:
                        server.stop()
                        server.start()
                        return _method_obj(*args, **kwargs)
                    elif e.code == ERROR_CODE_BASE - 2 and self.handlers['on']:  # Not Found
                        try:
                            self.handlers['on'] = False
                            # any handler returns True will break the left handlers
                            any(handler(self.handlers.get('device', None)) for handler in self.handlers['handlers'])
                        finally:
                            self.handlers['on'] = True
                        return _method_obj(*args, **kwargs)
                    raise
            return wrapper

        return JsonRPCClient(self.rpc_uri,
                             timeout=timeout,
                             method_class=_JsonRPCMethod)

    def __jsonrpc(self):
        return JsonRPCClient(self.rpc_uri, timeout=int(os.environ.get("JSONRPC_TIMEOUT", 90)))

    def sdk_version(self):
        '''sdk version of connected device.'''
        if self.__sdk == 0:
            try:
                self.__sdk = int(self.adb.cmd("shell", "getprop", "ro.build.version.sdk").communicate()[0].decode("utf-8").strip())
            except:
                pass
        return self.__sdk

    def start(self, timeout=5):
        if self.sdk_version() < 18:
            files = self.push()
            cmd = list(itertools.chain(
                ["shell", "uiautomator", "runtest"],
                files,
                ["-c", "com.github.uiautomatorstub.Stub"]
            ))
        else:
            self.install()
            cmd = ["shell", "am", "instrument", "-w",
                   "com.github.uiautomator.test/android.support.test.runner.AndroidJUnitRunner"]

        self.uiautomator_process = self.adb.cmd(*cmd)
        self.adb.forward(self.local_port, self.device_port)

        while not self.alive and timeout > 0:
            time.sleep(0.1)
            timeout -= 0.1
        if not self.alive:
            raise IOError("RPC server not started!")

    def ping(self):
        try:
            return self.__jsonrpc().ping()
        except:
            return None

    @property
    def alive(self):
        '''Check if the rpc server is alive.'''
        return self.ping() == "pong"

    def stop(self):
        '''Stop the rpc server.'''
        if self.uiautomator_process and self.uiautomator_process.poll() is None:
            res = None
            try:
                res = urllib2.urlopen(self.stop_uri)
                self.uiautomator_process.wait()
            except:
                self.uiautomator_process.kill()
            finally:
                if res is not None:
                    res.close()
                self.uiautomator_process = None
        try:
            out = self.adb.cmd("shell", "ps", "-C", "uiautomator").communicate()[0].decode("utf-8").strip().splitlines()
            if out:
                index = out[0].split().index("PID")
                for line in out[1:]:
                    if len(line.split()) > index:
                        self.adb.cmd("shell", "kill", "-9", line.split()[index]).wait()
        except:
            pass

    @property
    def stop_uri(self):
        return "http://%s:%d/stop" % (self.adb.adb_server_host, self.local_port)

    @property
    def rpc_uri(self):
        return "http://%s:%d/jsonrpc/0" % (self.adb.adb_server_host, self.local_port)

    @property
    def screenshot_uri(self):
        return "http://%s:%d/screenshot/0" % (self.adb.adb_server_host, self.local_port)

    def screenshot(self, filename=None, scale=1.0, quality=100):
        if self.sdk_version() >= 18:
            try:
                req = urllib2.Request("%s?scale=%f&quality=%f" % (self.screenshot_uri, scale, quality))
                result = urllib2.urlopen(req, timeout=30)
                if filename:
                    with open(filename, 'wb') as f:
                        f.write(result.read())
                        return filename
                else:
                    return result.read()
            except:
                pass
        return None


class AutomatorDevice(object):

    '''uiautomator wrapper of android device'''

    __orientation = (  # device orientation
        (0, "natural", "n", 0),
        (1, "left", "l", 90),
        (2, "upsidedown", "u", 180),
        (3, "right", "r", 270)
    )
    __alias = {
        "width": "displayWidth",
        "height": "displayHeight"
    }

    def __init__(self, serial=None, local_port=None, adb_server_host=None, adb_server_port=None):
        self.server = AutomatorServer(
            serial=serial,
            local_port=local_port,
            adb_server_host=adb_server_host,
            adb_server_port=adb_server_port
        )

    def __call__(self, **kwargs):
        return AutomatorDeviceObject(self, Selector(**kwargs))

    def __getattr__(self, attr):
        '''alias of fields in info property.'''
        info = self.info
        if attr in info:
            return info[attr]
        elif attr in self.__alias:
            return info[self.__alias[attr]]
        else:
            raise AttributeError("%s attribute not found!" % attr)

    @property
    def info(self):
        '''Get the device info.'''
        return self.server.jsonrpc.deviceInfo()

    def click(self, x, y):
        '''click at arbitrary coordinates.'''
        return self.server.jsonrpc.click(x, y)

    def long_click(self, x, y):
        '''long click at arbitrary coordinates.'''
        return self.swipe(x, y, x + 1, y + 1)

    def swipe(self, sx, sy, ex, ey, steps=100):
        return self.server.jsonrpc.swipe(sx, sy, ex, ey, steps)

    def swipePoints(self, points, steps=100):
        ppoints = []
        for p in points:
            ppoints.append(p[0])
            ppoints.append(p[1])
        return self.server.jsonrpc.swipePoints(ppoints, steps)

    def drag(self, sx, sy, ex, ey, steps=100):
        '''Swipe from one point to another point.'''
        return self.server.jsonrpc.drag(sx, sy, ex, ey, steps)

    def dump(self, filename=None, compressed=True, pretty=True):
        '''dump device window and pull to local file.'''
        content = self.server.jsonrpc.dumpWindowHierarchy(compressed, None)
        if filename:
            with open(filename, "wb") as f:
                f.write(content.encode("utf-8"))
        if pretty and "\n " not in content:
            xml_text = xml.dom.minidom.parseString(content.encode("utf-8"))
            content = U(xml_text.toprettyxml(indent='  '))
        return content

    def screenshot(self, filename, scale=1.0, quality=100):
        '''take screenshot.'''
        result = self.server.screenshot(filename, scale, quality)
        if result:
            return result

        device_file = self.server.jsonrpc.takeScreenshot("screenshot.png",
                                                         scale, quality)
        if not device_file:
            return None
        p = self.server.adb.cmd("pull", device_file, filename)
        p.wait()
        self.server.adb.cmd("shell", "rm", device_file).wait()
        return filename if p.returncode is 0 else None

    def freeze_rotation(self, freeze=True):
        '''freeze or unfreeze the device rotation in current status.'''
        self.server.jsonrpc.freezeRotation(freeze)

    @property
    def orientation(self):
        '''
        orienting the devie to left/right or natural.
        left/l:       rotation=90 , displayRotation=1
        right/r:      rotation=270, displayRotation=3
        natural/n:    rotation=0  , displayRotation=0
        upsidedown/u: rotation=180, displayRotation=2
        '''
        return self.__orientation[self.info["displayRotation"]][1]

    @orientation.setter
    def orientation(self, value):
        '''setter of orientation property.'''
        for values in self.__orientation:
            if value in values:
                # can not set upside-down until api level 18.
                self.server.jsonrpc.setOrientation(values[1])
                break
        else:
            raise ValueError("Invalid orientation.")

    @property
    def last_traversed_text(self):
        '''get last traversed text. used in webview for highlighted text.'''
        return self.server.jsonrpc.getLastTraversedText()

    def clear_traversed_text(self):
        '''clear the last traversed text.'''
        self.server.jsonrpc.clearLastTraversedText()

    @property
    def open(self):
        '''
        Open notification or quick settings.
        Usage:
        d.open.notification()
        d.open.quick_settings()
        '''
        @param_to_property(action=["notification", "quick_settings"])
        def _open(action):
            if action == "notification":
                return self.server.jsonrpc.openNotification()
            else:
                return self.server.jsonrpc.openQuickSettings()
        return _open

    @property
    def handlers(self):
        obj = self

        class Handlers(object):

            def on(self, fn):
                if fn not in obj.server.handlers['handlers']:
                    obj.server.handlers['handlers'].append(fn)
                obj.server.handlers['device'] = obj
                return fn

            def off(self, fn):
                if fn in obj.server.handlers['handlers']:
                    obj.server.handlers['handlers'].remove(fn)

        return Handlers()

    @property
    def watchers(self):
        obj = self

        class Watchers(list):

            def __init__(self):
                for watcher in obj.server.jsonrpc.getWatchers():
                    self.append(watcher)

            @property
            def triggered(self):
                return obj.server.jsonrpc.hasAnyWatcherTriggered()

            def remove(self, name=None):
                if name:
                    obj.server.jsonrpc.removeWatcher(name)
                else:
                    for name in self:
                        obj.server.jsonrpc.removeWatcher(name)

            def reset(self):
                obj.server.jsonrpc.resetWatcherTriggers()
                return self

            def run(self):
                obj.server.jsonrpc.runWatchers()
                return self
        return Watchers()

    def watcher(self, name):
        obj = self

        class Watcher(object):

            def __init__(self):
                self.__selectors = []

            @property
            def triggered(self):
                return obj.server.jsonrpc.hasWatcherTriggered(name)

            def remove(self):
                obj.server.jsonrpc.removeWatcher(name)

            def when(self, **kwargs):
                self.__selectors.append(Selector(**kwargs))
                return self

            def click(self, **kwargs):
                obj.server.jsonrpc.registerClickUiObjectWatcher(name, self.__selectors, Selector(**kwargs))

            @property
            def press(self):
                @param_to_property(
                    "home", "back", "left", "right", "up", "down", "center",
                    "search", "enter", "delete", "del", "recent", "volume_up",
                    "menu", "volume_down", "volume_mute", "camera", "power")
                def _press(*args):
                    obj.server.jsonrpc.registerPressKeyskWatcher(name, self.__selectors, args)
                return _press
        return Watcher()

    @property
    def press(self):
        '''
        press key via name or key code. Supported key name includes:
        home, back, left, right, up, down, center, menu, search, enter,
        delete(or del), recent(recent apps), volume_up, volume_down,
        volume_mute, camera, power.
        Usage:
        d.press.back()  # press back key
        d.press.menu()  # press home key
        d.press(89)     # press keycode
        '''
        @param_to_property(
            key=["home", "back", "left", "right", "up", "down", "center",
                 "menu", "search", "enter", "delete", "del", "recent",
                 "volume_up", "volume_down", "volume_mute", "camera", "power"]
        )
        def _press(key, meta=None):
            if isinstance(key, int):
                return self.server.jsonrpc.pressKeyCode(key, meta) if meta else self.server.jsonrpc.pressKeyCode(key)
            else:
                return self.server.jsonrpc.pressKey(str(key))
        return _press

    def wakeup(self):
        '''turn on screen in case of screen off.'''
        self.server.jsonrpc.wakeUp()

    def sleep(self):
        '''turn off screen in case of screen on.'''
        self.server.jsonrpc.sleep()

    @property
    def screen(self):
        '''
        Turn on/off screen.
        Usage:
        d.screen.on()
        d.screen.off()

        d.screen == 'on'  # Check if the screen is on, same as 'd.screenOn'
        d.screen == 'off'  # Check if the screen is off, same as 'not d.screenOn'
        '''
        devive_self = self

        class _Screen(object):
            def on(self):
                return devive_self.wakeup()

            def off(self):
                return devive_self.sleep()

            def __call__(self, action):
                if action == "on":
                    return self.on()
                elif action == "off":
                    return self.off()
                else:
                    raise AttributeError("Invalid parameter: %s" % action)

            def __eq__(self, value):
                info = devive_self.info
                if "screenOn" not in info:
                    raise EnvironmentError("Not supported on Android 4.3 and belows.")
                if value in ["on", "On", "ON"]:
                    return info["screenOn"]
                elif value in ["off", "Off", "OFF"]:
                    return not info["screenOn"]
                raise ValueError("Invalid parameter. It can only be compared with on/off.")

            def __ne__(self, value):
                return not self.__eq__(value)

        return _Screen()

    @property
    def wait(self):
        '''
        Waits for the current application to idle or window update event occurs.
        Usage:
        d.wait.idle(timeout=1000)
        d.wait.update(timeout=1000, package_name="com.android.settings")
        '''
        @param_to_property(action=["idle", "update"])
        def _wait(action, timeout=1000, package_name=None):
            if timeout / 1000 + 5 > int(os.environ.get("JSONRPC_TIMEOUT", 90)):
                http_timeout = timeout / 1000 + 5
            else:
                http_timeout = int(os.environ.get("JSONRPC_TIMEOUT", 90))
            if action == "idle":
                return self.server.jsonrpc_wrap(timeout=http_timeout).waitForIdle(timeout)
            elif action == "update":
                return self.server.jsonrpc_wrap(timeout=http_timeout).waitForWindowUpdate(package_name, timeout)
        return _wait

    def exists(self, **kwargs):
        '''Check if the specified ui object by kwargs exists.'''
        return self(**kwargs).exists

Device = AutomatorDevice


class AutomatorDeviceUiObject(object):

    '''Represent a UiObject, on which user can perform actions, such as click, set text
    '''

    __alias = {'description': "contentDescription"}

    def __init__(self, device, selector):
        self.device = device
        self.jsonrpc = device.server.jsonrpc
        self.selector = selector

    @property
    def exists(self):
        '''check if the object exists in current window.'''
        return self.jsonrpc.exist(self.selector)

    def __getattr__(self, attr):
        '''alias of fields in info property.'''
        info = self.info
        if attr in info:
            return info[attr]
        elif attr in self.__alias:
            return info[self.__alias[attr]]
        else:
            raise AttributeError("%s attribute not found!" % attr)

    @property
    def info(self):
        '''ui object info.'''
        return self.jsonrpc.objInfo(self.selector)

    def set_text(self, text):
        '''set the text field.'''
        if text in [None, ""]:
            return self.jsonrpc.clearTextField(self.selector)  # TODO no return
        else:
            return self.jsonrpc.setText(self.selector, text)

    def clear_text(self):
        '''clear text. alias for set_text(None).'''
        self.set_text(None)

    @property
    def click(self):
        '''
        click on the ui object.
        Usage:
        d(text="Clock").click()  # click on the center of the ui object
        d(text="OK").click.wait(timeout=3000) # click and wait for the new window update
        d(text="John").click.topleft() # click on the topleft of the ui object
        d(text="John").click.bottomright() # click on the bottomright of the ui object
        '''
        @param_to_property(action=["tl", "topleft", "br", "bottomright", "wait"])
        def _click(action=None, timeout=3000):
            if action is None:
                return self.jsonrpc.click(self.selector)
            elif action in ["tl", "topleft", "br", "bottomright"]:
                return self.jsonrpc.click(self.selector, action)
            else:
                return self.jsonrpc.clickAndWaitForNewWindow(self.selector, timeout)
        return _click

    @property
    def long_click(self):
        '''
        Perform a long click action on the object.
        Usage:
        d(text="Image").long_click()  # long click on the center of the ui object
        d(text="Image").long_click.topleft()  # long click on the topleft of the ui object
        d(text="Image").long_click.bottomright()  # long click on the topleft of the ui object
        '''
        @param_to_property(corner=["tl", "topleft", "br", "bottomright"])
        def _long_click(corner=None):
            info = self.info
            if info["longClickable"]:
                if corner:
                    return self.jsonrpc.longClick(self.selector, corner)
                else:
                    return self.jsonrpc.longClick(self.selector)
            else:
                bounds = info.get("visibleBounds") or info.get("bounds")
                if corner in ["tl", "topleft"]:
                    x = (5 * bounds["left"] + bounds["right"]) / 6
                    y = (5 * bounds["top"] + bounds["bottom"]) / 6
                elif corner in ["br", "bottomright"]:
                    x = (bounds["left"] + 5 * bounds["right"]) / 6
                    y = (bounds["top"] + 5 * bounds["bottom"]) / 6
                else:
                    x = (bounds["left"] + bounds["right"]) / 2
                    y = (bounds["top"] + bounds["bottom"]) / 2
                return self.device.long_click(x, y)
        return _long_click

    @property
    def drag(self):
        '''
        Drag the ui object to other point or ui object.
        Usage:
        d(text="Clock").drag.to(x=100, y=100)  # drag to point (x,y)
        d(text="Clock").drag.to(text="Remove") # drag to another object
        '''
        def to(obj, *args, **kwargs):
            if len(args) >= 2 or "x" in kwargs or "y" in kwargs:
                drag_to = lambda x, y, steps=100: self.jsonrpc.dragTo(self.selector, x, y, steps)
            else:
                drag_to = lambda steps=100, **kwargs: self.jsonrpc.dragTo(self.selector, Selector(**kwargs), steps)
            return drag_to(*args, **kwargs)
        return type("Drag", (object,), {"to": to})()

    def gesture(self, start1, start2, *args, **kwargs):
        '''
        perform two point gesture.
        Usage:
        d().gesture(startPoint1, startPoint2).to(endPoint1, endPoint2, steps)
        d().gesture(startPoint1, startPoint2, endPoint1, endPoint2, steps)
        '''
        def to(obj_self, end1, end2, steps=100):
            ctp = lambda pt: point(*pt) if type(pt) == tuple else pt  # convert tuple to point
            s1, s2, e1, e2 = ctp(start1), ctp(start2), ctp(end1), ctp(end2)
            return self.jsonrpc.gesture(self.selector, s1, s2, e1, e2, steps)
        obj = type("Gesture", (object,), {"to": to})()
        return obj if len(args) == 0 else to(None, *args, **kwargs)

    def gestureM(self, start1, start2, start3, *args, **kwargs):
        '''
        perform 3 point gesture.
        Usage:
        d().gestureM((100,200),(100,300),(100,400),(100,400),(100,400),(100,400))
        d().gestureM((100,200),(100,300),(100,400)).to((100,400),(100,400),(100,400))
        '''
        def to(obj_self, end1, end2, end3, steps=100):
            ctp = lambda pt: point(*pt) if type(pt) == tuple else pt  # convert tuple to point
            s1, s2, s3, e1, e2, e3 = ctp(start1), ctp(start2), ctp(start3), ctp(end1), ctp(end2), ctp(end3)
            return self.jsonrpc.gesture(self.selector, s1, s2, s3, e1, e2, e3, steps)
        obj = type("Gesture", (object,), {"to": to})()
        return obj if len(args) == 0 else to(None, *args, **kwargs)

    @property
    def pinch(self):
        '''
        Perform two point gesture from edge to center(in) or center to edge(out).
        Usages:
        d().pinch.In(percent=100, steps=10)
        d().pinch.Out(percent=100, steps=100)
        '''
        @param_to_property(in_or_out=["In", "Out"])
        def _pinch(in_or_out="Out", percent=100, steps=50):
            if in_or_out in ["Out", "out"]:
                return self.jsonrpc.pinchOut(self.selector, percent, steps)
            elif in_or_out in ["In", "in"]:
                return self.jsonrpc.pinchIn(self.selector, percent, steps)
        return _pinch

    @property
    def swipe(self):
        '''
        Perform swipe action. if device platform greater than API 18, percent can be used and value between 0 and 1
        Usages:
        d().swipe.right()
        d().swipe.left(steps=10)
        d().swipe.up(steps=10)
        d().swipe.down()
        d().swipe("right", steps=20)
        d().swipe("right", steps=20, percent=0.5)
        '''
        @param_to_property(direction=["up", "down", "right", "left"])
        def _swipe(direction="left", steps=10, percent=1):
            if percent == 1:
                return self.jsonrpc.swipe(self.selector, direction, steps)
            else:
                return self.jsonrpc.swipe(self.selector, direction, percent, steps)
        return _swipe

    @property
    def wait(self):
        '''
        Wait until the ui object gone or exist.
        Usage:
        d(text="Clock").wait.gone()  # wait until it's gone.
        d(text="Settings").wait.exists() # wait until it appears.
        '''
        @param_to_property(action=["exists", "gone"])
        def _wait(action, timeout=3000):
            if timeout / 1000 + 5 > int(os.environ.get("JSONRPC_TIMEOUT", 90)):
                http_timeout = timeout / 1000 + 5
            else:
                http_timeout = int(os.environ.get("JSONRPC_TIMEOUT", 90))
            method = self.device.server.jsonrpc_wrap(
                timeout=http_timeout
            ).waitUntilGone if action == "gone" else self.device.server.jsonrpc_wrap(timeout=http_timeout).waitForExists
            return method(self.selector, timeout)
        return _wait


class AutomatorDeviceNamedUiObject(AutomatorDeviceUiObject):

    def __init__(self, device, name):
        super(AutomatorDeviceNamedUiObject, self).__init__(device, name)

    def child(self, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.getChild(self.selector, Selector(**kwargs))
        )

    def sibling(self, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.getFromParent(self.selector, Selector(**kwargs))
        )


class AutomatorDeviceObject(AutomatorDeviceUiObject):

    '''Represent a generic UiObject/UiScrollable/UiCollection,
    on which user can perform actions, such as click, set text
    '''

    def __init__(self, device, selector):
        super(AutomatorDeviceObject, self).__init__(device, selector)

    def child(self, **kwargs):
        '''set childSelector.'''
        return AutomatorDeviceObject(
            self.device,
            self.selector.clone().child(**kwargs)
        )

    def sibling(self, **kwargs):
        '''set fromParent selector.'''
        return AutomatorDeviceObject(
            self.device,
            self.selector.clone().sibling(**kwargs)
        )

    child_selector, from_parent = child, sibling

    def child_by_text(self, txt, **kwargs):
        if "allow_scroll_search" in kwargs:
            allow_scroll_search = kwargs.pop("allow_scroll_search")
            name = self.jsonrpc.childByText(
                self.selector,
                Selector(**kwargs),
                txt,
                allow_scroll_search
            )
        else:
            name = self.jsonrpc.childByText(
                self.selector,
                Selector(**kwargs),
                txt
            )
        return AutomatorDeviceNamedUiObject(self.device, name)

    def child_by_description(self, txt, **kwargs):
        if "allow_scroll_search" in kwargs:
            allow_scroll_search = kwargs.pop("allow_scroll_search")
            name = self.jsonrpc.childByDescription(
                self.selector,
                Selector(**kwargs),
                txt,
                allow_scroll_search
            )
        else:
            name = self.jsonrpc.childByDescription(
                self.selector,
                Selector(**kwargs),
                txt
            )
        return AutomatorDeviceNamedUiObject(self.device, name)

    def child_by_instance(self, inst, **kwargs):
        return AutomatorDeviceNamedUiObject(
            self.device,
            self.jsonrpc.childByInstance(self.selector, Selector(**kwargs), inst)
        )

    @property
    def count(self):
        return self.jsonrpc.count(self.selector)

    def __len__(self):
        return self.count

    def __getitem__(self, index):
        count = self.count
        if index >= count:
            raise IndexError()
        elif count == 1:
            return self
        else:
            selector = self.selector.clone()
            selector["instance"] = index
            return AutomatorDeviceObject(self.device, selector)

    def __iter__(self):
        obj, length = self, self.count

        class Iter(object):

            def __init__(self):
                self.index = -1

            def next(self):
                self.index += 1
                if self.index < length:
                    return obj[self.index]
                else:
                    raise StopIteration()
            __next__ = next

        return Iter()

    def right(self, **kwargs):
        def onrightof(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect2["left"] - rect1["right"] if top < bottom else -1
        return self.__view_beside(onrightof, **kwargs)

    def left(self, **kwargs):
        def onleftof(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect1["left"] - rect2["right"] if top < bottom else -1
        return self.__view_beside(onleftof, **kwargs)

    def up(self, **kwargs):
        def above(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect1["top"] - rect2["bottom"] if left < right else -1
        return self.__view_beside(above, **kwargs)

    def down(self, **kwargs):
        def under(rect1, rect2):
            left, top, right, bottom = intersect(rect1, rect2)
            return rect2["top"] - rect1["bottom"] if left < right else -1
        return self.__view_beside(under, **kwargs)

    def __view_beside(self, onsideof, **kwargs):
        bounds = self.info["bounds"]
        min_dist, found = -1, None
        for ui in AutomatorDeviceObject(self.device, Selector(**kwargs)):
            dist = onsideof(bounds, ui.info["bounds"])
            if dist >= 0 and (min_dist < 0 or dist < min_dist):
                min_dist, found = dist, ui
        return found

    @property
    def fling(self):
        '''
        Perform fling action.
        Usage:
        d().fling()  # default vertically, forward
        d().fling.horiz.forward()
        d().fling.vert.backward()
        d().fling.toBeginning(max_swipes=100) # vertically
        d().fling.horiz.toEnd()
        '''
        @param_to_property(
            dimention=["vert", "vertically", "vertical", "horiz", "horizental", "horizentally"],
            action=["forward", "backward", "toBeginning", "toEnd"]
        )
        def _fling(dimention="vert", action="forward", max_swipes=1000):
            vertical = dimention in ["vert", "vertically", "vertical"]
            if action == "forward":
                return self.jsonrpc.flingForward(self.selector, vertical)
            elif action == "backward":
                return self.jsonrpc.flingBackward(self.selector, vertical)
            elif action == "toBeginning":
                return self.jsonrpc.flingToBeginning(self.selector, vertical, max_swipes)
            elif action == "toEnd":
                return self.jsonrpc.flingToEnd(self.selector, vertical, max_swipes)

        return _fling

    @property
    def scroll(self):
        '''
        Perfrom scroll action.
        Usage:
        d().scroll(steps=50) # default vertically and forward
        d().scroll.horiz.forward(steps=100)
        d().scroll.vert.backward(steps=100)
        d().scroll.horiz.toBeginning(steps=100, max_swipes=100)
        d().scroll.vert.toEnd(steps=100)
        d().scroll.horiz.to(text="Clock")
        '''
        def __scroll(vertical, forward, steps=100):
            method = self.jsonrpc.scrollForward if forward else self.jsonrpc.scrollBackward
            return method(self.selector, vertical, steps)

        def __scroll_to_beginning(vertical, steps=100, max_swipes=1000):
            return self.jsonrpc.scrollToBeginning(self.selector, vertical, max_swipes, steps)

        def __scroll_to_end(vertical, steps=100, max_swipes=1000):
            return self.jsonrpc.scrollToEnd(self.selector, vertical, max_swipes, steps)

        def __scroll_to(vertical, **kwargs):
            return self.jsonrpc.scrollTo(self.selector, Selector(**kwargs), vertical)

        @param_to_property(
            dimention=["vert", "vertically", "vertical", "horiz", "horizental", "horizentally"],
            action=["forward", "backward", "toBeginning", "toEnd", "to"])
        def _scroll(dimention="vert", action="forward", **kwargs):
            vertical = dimention in ["vert", "vertically", "vertical"]
            if action in ["forward", "backward"]:
                return __scroll(vertical, action == "forward", **kwargs)
            elif action == "toBeginning":
                return __scroll_to_beginning(vertical, **kwargs)
            elif action == "toEnd":
                return __scroll_to_end(vertical, **kwargs)
            elif action == "to":
                return __scroll_to(vertical, **kwargs)
        return _scroll

device = AutomatorDevice()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from mock import MagicMock, patch, call
from uiautomator import AutomatorServer, JsonRPCError


class TestAutomatorServer(unittest.TestCase):

    def setUp(self):
        self.Adb_patch = patch('uiautomator.Adb')
        self.Adb = self.Adb_patch.start()

    def tearDown(self):
        self.Adb.stop()

    def test_local_port(self):
        self.assertEqual(AutomatorServer("1234", 9010).local_port, 9010)
        self.Adb.assert_called_once_with(serial="1234", adb_server_host=None, adb_server_port=None)

    def test_local_port_forwarded(self):
        self.Adb.return_value.forward_list.return_value = [
            ("1234", "tcp:1001", "tcp:9009"),
            ("1234", "tcp:1000", "tcp:9008")
        ]
        self.Adb.return_value.device_serial.return_value = "1234"
        self.assertEqual(AutomatorServer("1234").local_port, 1000)

    def test_local_port_scanning(self):
        with patch('uiautomator.next_local_port') as next_local_port:
            self.Adb.return_value.forward_list.return_value = []
            next_local_port.return_value = 1234
            self.assertEqual(AutomatorServer("abcd", None).local_port,
                             next_local_port.return_value)

            next_local_port.return_value = 14321
            self.Adb.return_value.forward_list.return_value = Exception("error")
            self.assertEqual(AutomatorServer("abcd", None).local_port,
                             next_local_port.return_value)

    def test_device_port(self):
        self.assertEqual(AutomatorServer().device_port, 9008)

    def test_start_success(self):
        server = AutomatorServer()
        server.push = MagicMock()
        server.push.return_value = ["bundle.jar", "uiautomator-stub.jar"]
        server.ping = MagicMock()
        server.ping.return_value = "pong"
        server.adb = MagicMock()
        server.start()
        server.adb.cmd.assert_valled_onec_with('shell', 'uiautomator', 'runtest', 'bundle.jar', 'uiautomator-stub.jar', '-c', 'com.github.uiautomatorstub.Stub')

    def test_start_error(self):
        server = AutomatorServer()
        server.push = MagicMock()
        server.push.return_value = ["bundle.jar", "uiautomator-stub.jar"]
        server.ping = MagicMock()
        server.ping.return_value = None
        server.adb = MagicMock()
        with patch("time.sleep"):
            with self.assertRaises(IOError):
                server.start()

    def test_auto_start(self):
        try:
            import urllib2
        except ImportError:
            import urllib.request as urllib2
        with patch("uiautomator.JsonRPCMethod") as JsonRPCMethod:
            returns = [urllib2.URLError("error"), "ok"]
            def side_effect():
                result = returns.pop(0)
                if isinstance(result, Exception):
                    raise result
                return result
            JsonRPCMethod.return_value.side_effect = side_effect
            server = AutomatorServer()
            server.start = MagicMock()
            server.stop = MagicMock()
            self.assertEqual("ok", server.jsonrpc.any_method())
            server.start.assert_called_once_with(timeout=30)
        with patch("uiautomator.JsonRPCMethod") as JsonRPCMethod:
            returns = [JsonRPCError(-32000-1, "error msg"), "ok"]
            def side_effect():
                result = returns.pop(0)
                if isinstance(result, Exception):
                    raise result
                return result
            JsonRPCMethod.return_value.side_effect = side_effect
            server = AutomatorServer()
            server.start = MagicMock()
            server.stop = MagicMock()
            self.assertEqual("ok", server.jsonrpc.any_method())
            server.start.assert_called_once_with()
        with patch("uiautomator.JsonRPCMethod") as JsonRPCMethod:
            JsonRPCMethod.return_value.side_effect = JsonRPCError(-32000-2, "error msg")
            server = AutomatorServer()
            server.start = MagicMock()
            server.stop = MagicMock()
            with self.assertRaises(JsonRPCError):
                server.jsonrpc.any_method()

    def test_start_ping(self):
        with patch("uiautomator.JsonRPCClient") as JsonRPCClient:
            JsonRPCClient.return_value.ping.return_value = "pong"
            server = AutomatorServer()
            server.adb = MagicMock()
            server.adb.forward.return_value = 0
            self.assertEqual(server.ping(), "pong")

    def test_start_ping_none(self):
        with patch("uiautomator.JsonRPCClient") as JsonRPCClient:
            JsonRPCClient.return_value.ping.side_effect = Exception("error")
            server = AutomatorServer()
            self.assertEqual(server.ping(), None)


class TestAutomatorServer_Stop(unittest.TestCase):

    def setUp(self):
        try:
            import urllib2
            self.urlopen_patch = patch('urllib2.urlopen')
        except:
            self.urlopen_patch = patch('urllib.request.urlopen')
        finally:
            self.urlopen = self.urlopen_patch.start()

    def tearDown(self):
        self.urlopen_patch.stop()

    def test_screenshot(self):
        server = AutomatorServer()
        server.sdk_version = MagicMock()
        server.sdk_version.return_value = 17
        self.assertEqual(server.screenshot(), None)

        server.sdk_version.return_value = 18
        self.urlopen.return_value.read = MagicMock()
        self.urlopen.return_value.read.return_value = b"123456"
        self.assertEqual(server.screenshot(), b"123456")
        self.assertEqual(server.screenshot("/tmp/test.txt"), "/tmp/test.txt")

    def test_push(self):
        jars = ["bundle.jar", "uiautomator-stub.jar"]
        server = AutomatorServer()
        server.adb = MagicMock()
        self.assertEqual(set(server.push()), set(jars))
        for args in server.adb.cmd.call_args_list:
            self.assertEqual(args[0][0], "push")
            self.assertEqual(args[0][2], "/data/local/tmp/")

    def test_stop_started_server(self):
        server = AutomatorServer()
        server.adb = MagicMock()
        server.uiautomator_process = process = MagicMock()
        process.poll.return_value = None
        server.stop()
        process.wait.assert_called_once_with()

        server.uiautomator_process = process = MagicMock()
        process.poll.return_value = None
        self.urlopen.side_effect = IOError("error")
        server.stop()
        process.kill.assert_called_once_with()

    def test_stop(self):
        results = [
            b"USER     PID   PPID  VSIZE  RSS     WCHAN    PC         NAME\n\rsystem    372   126   635596 104808 ffffffff 00000000 S uiautomator",
            b"USER     PID   PPID  VSIZE  RSS     WCHAN    PC         NAME\r\nsystem    372   126   635596 104808 ffffffff 00000000 S uiautomator",
            b"USER     PID   PPID  VSIZE  RSS     WCHAN    PC         NAME\nsystem    372   126   635596 104808 ffffffff 00000000 S uiautomator",
            b"USER     PID   PPID  VSIZE  RSS     WCHAN    PC         NAME\rsystem    372   126   635596 104808 ffffffff 00000000 S uiautomator"
        ]
        for r in results:
            server = AutomatorServer()
            server.adb = MagicMock()
            server.adb.cmd.return_value.communicate.return_value = (r, "")
            server.stop()
            self.assertEqual(server.adb.cmd.call_args_list,
                             [call("shell", "ps", "-C", "uiautomator"), call("shell", "kill", "-9", "372")])


class TestJsonRPCError(unittest.TestCase):

    def testJsonRPCError(self):
        e = JsonRPCError(200, "error")
        self.assertEqual(200, e.code)
        self.assertTrue(len(str(e)) > 0)
        e = JsonRPCError("200", "error")
        self.assertEqual(200, e.code)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from mock import MagicMock, call
from uiautomator import AutomatorDeviceObject, Selector, AutomatorDeviceNamedUiObject


class TestDeviceObjInit(unittest.TestCase):

    def setUp(self):
        self.device = MagicMock()
        self.device.server.jsonrpc = MagicMock()

    def test_init(self):
        kwargs = {"text": "text", "className": "android"}
        self.device_obj = AutomatorDeviceObject(self.device,
                                                Selector(**kwargs))
        self.assertEqual(self.device_obj.selector,
                         Selector(**kwargs))
        self.assertEqual(self.device_obj.jsonrpc,
                         self.device.server.jsonrpc)


class TestDeviceObj(unittest.TestCase):

    def setUp(self):
        self.device = MagicMock()
        self.jsonrpc = self.device.server.jsonrpc = MagicMock()
        self.jsonrpc_wrap = self.device.server.jsonrpc_wrap = MagicMock()
        self.kwargs = {"text": "text", "className": "android"}
        self.obj = AutomatorDeviceObject(self.device,
                                         Selector(**self.kwargs))

    def test_child_selector(self):
        kwargs = {"text": "child text", "className": "android"}
        obj = self.obj.child_selector(**kwargs)
        self.assertEqual(len(obj.selector['childOrSibling']), 1)
        self.assertEqual(obj.selector['childOrSibling'][0], 'child')
        self.assertEqual(len(obj.selector['childOrSiblingSelector']), 1)
        self.assertEqual(obj.selector['childOrSiblingSelector'][0], Selector(**kwargs))

    def test_from_parent(self):
        kwargs = {"text": "parent text", "className": "android"}
        obj = self.obj.from_parent(**kwargs)
        self.assertEqual(len(obj.selector['childOrSibling']), 1)
        self.assertEqual(obj.selector['childOrSibling'][0], 'sibling')
        self.assertEqual(len(obj.selector['childOrSiblingSelector']), 1)
        self.assertEqual(obj.selector['childOrSiblingSelector'][0], Selector(**kwargs))

    def test_exists(self):
        self.jsonrpc.exist = MagicMock()
        self.jsonrpc.exist.return_value = True
        self.assertTrue(self.obj.exists)

        self.jsonrpc.exist.return_value = False
        self.assertFalse(self.obj.exists)

        self.assertEqual(self.jsonrpc.exist.call_args_list,
                         [call(self.obj.selector),
                          call(self.obj.selector)])

    def test_info(self):
        info = {"text": "item text"}
        self.jsonrpc.objInfo.return_value = info
        self.assertEqual(self.obj.info,
                         info)
        self.jsonrpc.objInfo.assert_called_once_with(self.obj.selector)

    def test_info_attr(self):
        info = {'contentDescription': '',
                'checked': False,
                'scrollable': False,
                'text': '',
                'packageName': 'android',
                'selected': False,
                'enabled': True,
                'bounds': {'top': 0,
                           'left': 0,
                           'right': 720,
                           'bottom': 1184},
                'className':
                'android.widget.FrameLayout',
                'focusable': False,
                'focused': False,
                'clickable': False,
                'checkable': False,
                'chileCount': 2,
                'longClickable': False,
                'visibleBounds': {'top': 0,
                                  'left': 0,
                                  'right': 720,
                                  'bottom': 1184}}
        self.jsonrpc.objInfo.return_value = info
        self.assertEqual(self.obj.info, info)
        self.jsonrpc.objInfo.assert_called_once_with(self.obj.selector)
        self.assertEqual(self.obj.description, info["contentDescription"])
        for k in info:
            self.assertEqual(getattr(self.obj, k), info[k])

        with self.assertRaises(AttributeError):
            self.obj.not_exists

    def test_text(self):
        self.jsonrpc.clearTextField = MagicMock()
        self.obj.set_text(None)
        self.obj.set_text("")
        self.obj.clear_text()
        self.assertEqual(self.jsonrpc.clearTextField.call_args_list,
                         [call(self.obj.selector), call(self.obj.selector), call(self.obj.selector)])

        self.jsonrpc.setText.return_value = False
        texts = ["abc", "123", "()#*$&"]
        for text in texts:
            self.assertFalse(self.obj.set_text(text))
        self.assertEqual(self.jsonrpc.setText.call_args_list,
                         [call(self.obj.selector, t) for t in texts])

    def test_click(self):
        self.jsonrpc.click.return_value = False
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertFalse(self.obj.click(c))
        self.assertEqual(self.jsonrpc.click.call_args_list,
                         [call(self.obj.selector, c) for c in corners])

        self.jsonrpc.click = MagicMock()
        self.jsonrpc.click.return_value = True
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertTrue(getattr(self.obj.click, c)())
        self.assertEqual(self.jsonrpc.click.call_args_list,
                         [call(self.obj.selector, c) for c in corners])

        self.jsonrpc.click = MagicMock()
        self.jsonrpc.click.return_value = True
        self.assertTrue(self.obj.click())
        self.jsonrpc.click.assert_called_once_with(self.obj.selector)

    def test_click_wait(self):
        self.jsonrpc.clickAndWaitForNewWindow.return_value = True
        self.assertTrue(self.obj.click.wait(timeout=321))
        self.jsonrpc.clickAndWaitForNewWindow.assert_called_once_with(self.obj.selector, 321)

    def test_long_click(self):
        self.jsonrpc.longClick.return_value = False
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertFalse(self.obj.long_click(c))
        self.assertEqual(self.jsonrpc.longClick.call_args_list,
                         [call(self.obj.selector, c) for c in corners])

        self.jsonrpc.longClick = MagicMock()
        self.jsonrpc.longClick.return_value = True
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertTrue(getattr(self.obj.long_click, c)())
        self.assertEqual(self.jsonrpc.longClick.call_args_list,
                         [call(self.obj.selector, c) for c in corners])

        self.jsonrpc.longClick = MagicMock()
        self.jsonrpc.longClick.return_value = True
        self.assertTrue(self.obj.long_click())
        self.jsonrpc.longClick.assert_called_once_with(self.obj.selector)

    def test_long_click_using_swipe(self):
        self.device.long_click.return_value = False
        self.jsonrpc.objInfo.return_value = {
            'longClickable': False,
            'visibleBounds': {
                'top': 0,
                'bottom': 60,
                'left': 0,
                'right': 60
            }
        }
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertFalse(self.obj.long_click(c))
        self.assertEqual(self.device.long_click.call_args_list,
                         [call(10, 10), call(10, 10), call(50, 50), call(50, 50)])

        self.device.long_click = MagicMock()
        self.device.long_click.return_value = True
        corners = ["tl", "topleft", "br", "bottomright"]
        for c in corners:
            self.assertTrue(getattr(self.obj.long_click, c)())
        self.assertEqual(self.device.long_click.call_args_list,
                         [call(10, 10), call(10, 10), call(50, 50), call(50, 50)])

        self.device.long_click = MagicMock()
        self.device.long_click.return_value = True
        self.assertTrue(self.obj.long_click())
        self.device.long_click.assert_called_once_with(30, 30)

    def test_drag_to(self):
        self.jsonrpc.dragTo.return_value = False
        self.assertFalse(self.obj.drag.to(10, 20, steps=10))
        self.jsonrpc.dragTo.return_value = True
        self.assertTrue(self.obj.drag.to(x=10, y=20, steps=20))

        sel = {"text": "text..."}
        self.assertTrue(self.obj.drag.to(steps=30, **sel))
        self.assertEqual(self.jsonrpc.dragTo.call_args_list,
                         [call(self.obj.selector, 10, 20, 10),
                          call(self.obj.selector, 10, 20, 20),
                          call(self.obj.selector, Selector(**sel), 30)])

    def test_gesture(self):
        self.jsonrpc.gesture.return_value = True
        self.assertTrue(self.obj.gesture((1, 1), (2, 2), (3, 3), (4, 4), 100))
        self.assertTrue(self.obj.gesture(4, 3).to(2, 1, 20))
        self.assertEqual(self.jsonrpc.gesture.call_args_list,
                         [call(self.obj.selector, {'x':1, 'y': 1}, {'x':2, 'y':2}, {'x':3, 'y':3}, {'x':4, 'y':4}, 100), call(self.obj.selector, 4, 3, 2, 1, 20)])

    def test_pinch(self):
        self.jsonrpc.pinchIn.return_value = True
        self.assertTrue(self.obj.pinch.In(percent=90, steps=30))
        self.assertTrue(self.obj.pinch("in", 80, 40))
        self.assertTrue(self.obj.pinch("In", 70, 50))
        self.assertEqual(self.jsonrpc.pinchIn.call_args_list,
                         [call(self.obj.selector, 90, 30), call(self.obj.selector, 80, 40), call(self.obj.selector, 70, 50)])

        self.jsonrpc.pinchOut.return_value = True
        self.assertTrue(self.obj.pinch.Out(percent=90, steps=30))
        self.assertTrue(self.obj.pinch("out", 80, 40))
        self.assertTrue(self.obj.pinch("Out", 70, 50))
        self.assertEqual(self.jsonrpc.pinchIn.call_args_list,
                         [call(self.obj.selector, 90, 30), call(self.obj.selector, 80, 40), call(self.obj.selector, 70, 50)])

    def test_swipe(self):
        self.jsonrpc.swipe.return_value = True
        dirs = ["up", "down", "right", "left"]
        for d in dirs:
            self.assertTrue(self.obj.swipe(d, 30))
        self.assertEqual(self.jsonrpc.swipe.call_args_list,
                         [call(self.obj.selector, d, 30) for d in dirs])

        self.jsonrpc.swipe = MagicMock()
        self.jsonrpc.swipe.return_value = True
        dirs = ["up", "down", "right", "left"]
        for d in dirs:
            self.assertTrue(getattr(self.obj.swipe, d)(steps=30))
        self.assertEqual(self.jsonrpc.swipe.call_args_list,
                         [call(self.obj.selector, d, 30) for d in dirs])

    def test_fling(self):
        self.jsonrpc.flingForward.return_value = True
        self.assertTrue(self.obj.fling.horiz.forward())
        self.assertTrue(self.obj.fling.horizentally.forward())
        self.assertTrue(self.obj.fling.vert.forward())
        self.assertTrue(self.obj.fling())
        self.assertEqual(self.jsonrpc.flingForward.call_args_list,
                         [call(self.obj.selector, False), call(self.obj.selector, False), call(self.obj.selector, True), call(self.obj.selector, True)])

        self.jsonrpc.flingBackward.return_value = True
        self.assertTrue(self.obj.fling.horiz.backward())
        self.assertTrue(self.obj.fling.horizentally.backward())
        self.assertTrue(self.obj.fling.vert.backward())
        self.assertTrue(self.obj.fling.vertically.backward())
        self.assertEqual(self.jsonrpc.flingBackward.call_args_list,
                         [call(self.obj.selector, False), call(self.obj.selector, False), call(self.obj.selector, True), call(self.obj.selector, True)])

        max_swipes = 1000
        self.jsonrpc.flingToBeginning.return_value = True
        self.assertTrue(self.obj.fling.horiz.toBeginning())
        self.assertTrue(self.obj.fling.horizentally.toBeginning())
        self.assertTrue(self.obj.fling.vert.toBeginning())
        self.assertTrue(self.obj.fling.vertically.toBeginning(max_swipes=100))
        self.assertEqual(self.jsonrpc.flingToBeginning.call_args_list,
                         [call(self.obj.selector, False, max_swipes), call(self.obj.selector, False, max_swipes), call(self.obj.selector, True, max_swipes), call(self.obj.selector, True, 100)])

        self.jsonrpc.flingToEnd.return_value = True
        self.assertTrue(self.obj.fling.horiz.toEnd())
        self.assertTrue(self.obj.fling.horizentally.toEnd())
        self.assertTrue(self.obj.fling.vert.toEnd())
        self.assertTrue(self.obj.fling.vertically.toEnd(max_swipes=100))
        self.assertEqual(self.jsonrpc.flingToEnd.call_args_list,
                         [call(self.obj.selector, False, max_swipes), call(self.obj.selector, False, max_swipes), call(self.obj.selector, True, max_swipes), call(self.obj.selector, True, 100)])

    def test_scroll(self):
        steps = 100
        max_swipes = 1000
        self.jsonrpc.scrollForward.return_value = True
        self.assertTrue(self.obj.scroll.horiz.forward())
        self.assertTrue(self.obj.scroll.horizentally.forward())
        self.assertTrue(self.obj.scroll.vert.forward())
        self.assertTrue(self.obj.scroll(steps=20))
        self.assertEqual(self.jsonrpc.scrollForward.call_args_list,
                         [call(self.obj.selector, False, steps), call(self.obj.selector, False, steps), call(self.obj.selector, True, steps), call(self.obj.selector, True, 20)])

        self.jsonrpc.scrollBackward.return_value = True
        self.assertTrue(self.obj.scroll.horiz.backward())
        self.assertTrue(self.obj.scroll.horizentally.backward())
        self.assertTrue(self.obj.scroll.vert.backward())
        self.assertTrue(self.obj.scroll.vertically.backward(steps=20))
        self.assertEqual(self.jsonrpc.scrollBackward.call_args_list,
                         [call(self.obj.selector, False, steps), call(self.obj.selector, False, steps), call(self.obj.selector, True, steps), call(self.obj.selector, True, 20)])

        self.jsonrpc.scrollToBeginning.return_value = True
        self.assertTrue(self.obj.scroll.horiz.toBeginning())
        self.assertTrue(self.obj.scroll.horizentally.toBeginning())
        self.assertTrue(self.obj.scroll.vert.toBeginning())
        self.assertTrue(self.obj.scroll.vertically.toBeginning(steps=20, max_swipes=100))
        self.assertEqual(self.jsonrpc.scrollToBeginning.call_args_list,
                         [call(self.obj.selector, False, max_swipes, steps), call(self.obj.selector, False, max_swipes, steps), call(self.obj.selector, True, max_swipes, steps), call(self.obj.selector, True, 100, 20)])

        self.jsonrpc.scrollToEnd.return_value = True
        self.assertTrue(self.obj.scroll.horiz.toEnd())
        self.assertTrue(self.obj.scroll.horizentally.toEnd())
        self.assertTrue(self.obj.scroll.vert.toEnd())
        self.assertTrue(self.obj.scroll.vertically.toEnd(steps=20, max_swipes=100))
        self.assertEqual(self.jsonrpc.scrollToEnd.call_args_list,
                         [call(self.obj.selector, False, max_swipes, steps), call(self.obj.selector, False, max_swipes, steps), call(self.obj.selector, True, max_swipes, steps), call(self.obj.selector, True, 100, 20)])

        info = {"text": "..."}
        self.jsonrpc.scrollTo.return_value = True
        self.assertTrue(self.obj.scroll.horiz.to(**info))
        self.assertTrue(self.obj.scroll.horizentally.to(**info))
        self.assertTrue(self.obj.scroll.vert.to(**info))
        self.assertTrue(self.obj.scroll.vertically.to(**info))
        self.assertEqual(self.jsonrpc.scrollTo.call_args_list,
                         [call(self.obj.selector, Selector(**info), False), call(self.obj.selector, Selector(**info), False), call(self.obj.selector, Selector(**info), True), call(self.obj.selector, Selector(**info), True)])

    def test_wait(self):
        timeout = 3000
        self.jsonrpc_wrap.return_value.waitUntilGone.return_value = True
        self.assertTrue(self.obj.wait.gone())
        self.jsonrpc_wrap.return_value.waitUntilGone.assert_called_once_with(self.obj.selector, timeout)
        self.jsonrpc_wrap.return_value.waitForExists.return_value = True
        self.assertTrue(self.obj.wait.exists(timeout=10))
        self.jsonrpc_wrap.return_value.waitForExists.assert_called_once_with(self.obj.selector, 10)

    def test_child_by_text(self):
        self.jsonrpc.childByText.return_value = "myname"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child_by_text("child text", **kwargs)
        self.jsonrpc.childByText.assert_called_once_with(Selector(**self.kwargs), Selector(**kwargs), "child text")
        self.assertEqual("myname", generic_obj.selector)

    def test_child_by_text_allow_scroll_search(self):
        self.jsonrpc.childByText.return_value = "myname"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child_by_text("child text", allow_scroll_search=False, **kwargs)
        self.jsonrpc.childByText.assert_called_once_with(
            Selector(**self.kwargs), Selector(**kwargs), "child text", False)
        self.assertEqual("myname", generic_obj.selector)

    def test_child_by_description(self):
        self.jsonrpc.childByDescription.return_value = "myname"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child_by_description("child text", **kwargs)
        self.jsonrpc.childByDescription.assert_called_once_with(
            Selector(**self.kwargs), Selector(**kwargs), "child text")
        self.assertEqual("myname", generic_obj.selector)

    def test_child_by_description_allow_scroll_search(self):
        self.jsonrpc.childByDescription.return_value = "myname"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child_by_description("child text", allow_scroll_search=False, **kwargs)
        self.jsonrpc.childByDescription.assert_called_once_with(
            Selector(**self.kwargs), Selector(**kwargs), "child text", False)
        self.assertEqual("myname", generic_obj.selector)

    def test_child_by_instance(self):
        self.jsonrpc.childByInstance.return_value = "myname"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child_by_instance(1234, **kwargs)
        self.jsonrpc.childByInstance.assert_called_once_with(Selector(**self.kwargs), Selector(**kwargs), 1234)
        self.assertEqual("myname", generic_obj.selector)

    def test_count(self):
        self.jsonrpc.count.return_value = 10
        self.assertEqual(self.obj.count, 10)
        self.jsonrpc.count.assert_called_once_with(Selector(**self.kwargs))

    def test_len(self):
        self.jsonrpc.count.return_value = 10
        self.assertEqual(len(self.obj), 10)

    def test_instance_list(self):
        count = 10
        self.jsonrpc.count.return_value = count
        for i in range(count):
            self.assertEqual(self.obj[i].selector["instance"], i)
        with self.assertRaises(IndexError):
            self.obj[count]
        self.jsonrpc.count.return_value = 1
        self.assertEqual(self.obj[0], self.obj)

    def test_instance_iter(self):
        count = 10
        self.jsonrpc.count.return_value = count
        for index, inst in enumerate(self.obj):
            self.assertEqual(inst.selector["instance"], index)

    def test_left(self):
        self.jsonrpc.objInfo.side_effect = [
            {"bounds": {'top': 200, 'bottom': 250, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 150, 'right': 200}},
            {"bounds": {'top': 200, 'bottom': 300, 'left': 150, 'right': 200}},
            {"bounds": {'top': 200, 'bottom': 300, 'left': 50, 'right': 100}}
        ]
        self.jsonrpc.count.return_value = 3
        self.assertEqual(self.obj.left().selector["instance"], 2)

    def test_right(self):
        self.jsonrpc.objInfo.side_effect = [
            {"bounds": {'top': 200, 'bottom': 250, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 150, 'right': 200}},
            {"bounds": {'top': 200, 'bottom': 300, 'left': 50, 'right': 100}},
            {"bounds": {'top': 200, 'bottom': 300, 'left': 150, 'right': 200}}
        ]
        self.jsonrpc.count.return_value = 3
        self.assertEqual(self.obj.right().selector["instance"], 2)

    def test_up(self):
        self.jsonrpc.objInfo.side_effect = [
            {"bounds": {'top': 200, 'bottom': 250, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 100, 'right': 150}},
            {"bounds": {'top': 150, 'bottom': 200, 'left': 150, 'right': 200}},
            {"bounds": {'top': 150, 'bottom': 200, 'left': 100, 'right': 200}}
        ]
        self.jsonrpc.count.return_value = 3
        self.assertEqual(self.obj.up().selector["instance"], 2)

    def test_down(self):
        self.jsonrpc.objInfo.side_effect = [
            {"bounds": {'top': 200, 'bottom': 250, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 150, 'right': 200}},
            {"bounds": {'top': 150, 'bottom': 200, 'left': 150, 'right': 200}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 100, 'right': 150}}
        ]
        self.jsonrpc.count.return_value = 3
        self.assertEqual(self.obj.down().selector["instance"], 2)

    def test_multiple_matched_down(self):
        self.jsonrpc.objInfo.side_effect = [
            {"bounds": {'top': 200, 'bottom': 250, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 300, 'left': 150, 'right': 200}},
            {"bounds": {'top': 150, 'bottom': 200, 'left': 150, 'right': 200}},
            {"bounds": {'top': 275, 'bottom': 300, 'left': 100, 'right': 150}},
            {"bounds": {'top': 300, 'bottom': 350, 'left': 100, 'right': 150}},
            {"bounds": {'top': 250, 'bottom': 275, 'left': 100, 'right': 150}}
        ]
        self.jsonrpc.count.return_value = 5
        self.assertEqual(self.obj.down().selector["instance"], 4)

class TestAutomatorDeviceNamedUiObject(unittest.TestCase):

    def setUp(self):
        self.device = MagicMock()
        self.jsonrpc = self.device.server.jsonrpc = MagicMock()
        self.name = "my-name"
        self.obj = AutomatorDeviceNamedUiObject(self.device, self.name)

    def test_child(self):
        self.jsonrpc.getChild.return_value = "another-name"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.child(**kwargs)
        self.jsonrpc.getChild.assert_called_once_with(self.name, Selector(**kwargs))
        self.assertEqual(generic_obj.selector, self.jsonrpc.getChild.return_value)

    def test_sibling(self):
        self.jsonrpc.getFromParent.return_value = "another-name"
        kwargs = {"className": "android", "text": "patern match text"}
        generic_obj = self.obj.sibling(**kwargs)
        self.jsonrpc.getFromParent.assert_called_once_with(self.name, Selector(**kwargs))
        self.assertEqual(generic_obj.selector, self.jsonrpc.getFromParent.return_value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from uiautomator import param_to_property


class TestParamToProperty(unittest.TestCase):

    def test_props(self):
        args_value = None
        kwargs_value = None

        @param_to_property("one", "two", "three")
        def func(*args, **kwargs):
            self.assertEqual(args, args_value)
            self.assertEqual(kwargs, kwargs_value)

        args_value = ("one", "two", "three")
        kwargs_value = {"test": 1}
        func.one.two.three(test=1)
        args_value = ("one", "three")
        kwargs_value = {"another_test": 100}
        func.one("three", another_test=100)
        args_value = ("one", "two", "three")
        kwargs_value = {}
        func("one", "two", "three")
        args_value = ("three", "one", "two")
        kwargs_value = {}
        func.three("one", "two")

    def test_kwprops(self):
        args_value = None
        kwargs_value = None

        @param_to_property(key=["one", "two", "three"])
        def func(*args, **kwargs):
            self.assertEqual(args, args_value)
            self.assertEqual(kwargs, kwargs_value)

        args_value = (1,)
        kwargs_value = {"key": "one"}
        func.one(1)
        args_value = (2, 3)
        kwargs_value = {"key": "two"}
        func.two(2, 3)
        args_value = ()
        kwargs_value = {}
        func()

    def test_error(self):
        @param_to_property(key=["one", "two", "three"])
        def func(*args, **kwargs):
            pass

        with self.assertRaises(AttributeError):
            func.one.one

        with self.assertRaises(SyntaxError):
            @param_to_property("a", "b", key=["one", "two", "three"])
            def func(*args, **kwargs):
                pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from mock import MagicMock, patch
import os
import subprocess
from uiautomator import Adb


class TestAdb(unittest.TestCase):

    def setUp(self):
        self.os_name = os.name

    def tearDown(self):
        os.name = self.os_name

    def test_serial(self):
        serial = "abcdef1234567890"
        adb = Adb(serial)
        self.assertEqual(adb.default_serial, serial)

        adb.devices = MagicMock()
        adb.devices.return_value = [serial, "123456"]
        self.assertEqual(adb.device_serial(), serial)

    def test_adb_from_env(self):
        home_dir = '/android/home'
        with patch.dict('os.environ', {'ANDROID_HOME': home_dir}):
            with patch('os.path.exists') as exists:
                exists.return_value = True

                os.name = "posix"  # linux
                adb_obj = Adb()
                adb_path = os.path.join(home_dir, "platform-tools", "adb")
                self.assertEqual(adb_obj.adb(), adb_path)
                exists.assert_called_once_with(adb_path)
                self.assertEqual(adb_obj.adb(), adb_path)
                # the second call will return the __adb_cmd directly
                exists.assert_called_once_with(adb_path)

                os.name = "nt"  # linux
                adb_obj = Adb()
                adb_path = os.path.join(home_dir, "platform-tools", "adb.exe")
                self.assertEqual(adb_obj.adb(), adb_path)

                exists.return_value = False
                with self.assertRaises(EnvironmentError):
                    Adb().adb()

    def test_adb_from_find(self):
        with patch.dict('os.environ', {}, clear=True):
            with patch("distutils.spawn.find_executable") as find_executable:
                find_executable.return_value = "/usr/bin/adb"
                with patch("os.path.realpath") as realpath:
                    realpath.return_value = "/home/user/android/platform-tools/adb"
                    self.assertEqual(realpath.return_value, Adb().adb())
                    find_executable.assert_called_once_with("adb")  # find_exectable should be called once
                    realpath.assert_called_once_with(find_executable.return_value)
                    realpath.return_value = find_executable.return_value
                    self.assertEqual(find_executable.return_value, Adb().adb())
                find_executable.return_value = None
                call_count = find_executable.call_count
                with self.assertRaises(EnvironmentError):
                    Adb().adb()
                self.assertEqual(call_count + 1, find_executable.call_count)

    def test_devices(self):
        adb = Adb()
        adb.raw_cmd = MagicMock()
        adb.raw_cmd.return_value.communicate.return_value = (b"List of devices attached \r\n014E05DE0F02000E\tdevice\r\n489328DKFL7DF\tdevice", b"")
        self.assertEqual(adb.devices(), {"014E05DE0F02000E": "device", "489328DKFL7DF": "device"})
        adb.raw_cmd.assert_called_once_with("devices")
        adb.raw_cmd.return_value.communicate.return_value = (b"List of devices attached \n\r014E05DE0F02000E\tdevice\n\r489328DKFL7DF\tdevice", b"")
        self.assertEqual(adb.devices(), {"014E05DE0F02000E": "device", "489328DKFL7DF": "device"})
        adb.raw_cmd.return_value.communicate.return_value = (b"List of devices attached \r014E05DE0F02000E\tdevice\r489328DKFL7DF\tdevice", b"")
        self.assertEqual(adb.devices(), {"014E05DE0F02000E": "device", "489328DKFL7DF": "device"})
        adb.raw_cmd.return_value.communicate.return_value = (b"List of devices attached \n014E05DE0F02000E\tdevice\n489328DKFL7DF\tdevice", b"")
        self.assertEqual(adb.devices(), {"014E05DE0F02000E": "device", "489328DKFL7DF": "device"})
        adb.raw_cmd.return_value.communicate.return_value = (b"not match", "")
        with self.assertRaises(EnvironmentError):
            adb.devices()

    def test_forward(self):
        adb = Adb()
        adb.cmd = MagicMock()
        adb.forward(90, 91)
        adb.cmd.assert_called_once_with("forward", "tcp:90", "tcp:91")
        adb.cmd.return_value.wait.assert_called_once_with()

    def test_adb_raw_cmd(self):
        import subprocess
        adb = Adb()
        adb.adb = MagicMock()
        adb.adb.return_value = "adb"
        args = ["a", "b", "c"]
        with patch("subprocess.Popen") as Popen:
            os.name = "posix"
            adb.raw_cmd(*args)
            Popen.assert_called_once_with(["%s %s" % (adb.adb(), " ".join(args))], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        with patch("subprocess.Popen") as Popen:
            os.name = "nt"
            adb.raw_cmd(*args)
            Popen.assert_called_once_with([adb.adb()] + list(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def test_adb_cmd(self):
        adb = Adb()
        adb.device_serial = MagicMock()
        adb.device_serial.return_value = "ANDROID_SERIAL"
        adb.raw_cmd = MagicMock()
        args = ["a", "b", "c"]
        adb.cmd(*args)
        adb.raw_cmd.assert_called_once_with("-s", "%s" % adb.device_serial(), *args)

        adb.device_serial.return_value = "ANDROID SERIAL"
        adb.raw_cmd = MagicMock()
        args = ["a", "b", "c"]
        adb.cmd(*args)
        adb.raw_cmd.assert_called_once_with("-s", "'%s'" % adb.device_serial(), *args)

    def test_adb_cmd_server_host(self):
        adb = Adb(adb_server_host="localhost", adb_server_port=5037)
        adb.adb = MagicMock()
        adb.adb.return_value = "adb"
        adb.device_serial = MagicMock()
        adb.device_serial.return_value = "ANDROID_SERIAL"
        args = ["a", "b", "c"]
        with patch("subprocess.Popen") as Popen:
            os.name = "nt"
            adb.raw_cmd(*args)
            Popen.assert_called_once_with(
                [adb.adb()] + args,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )

        adb = Adb(adb_server_host="test.com", adb_server_port=1000)
        adb.adb = MagicMock()
        adb.adb.return_value = "adb"
        adb.device_serial = MagicMock()
        adb.device_serial.return_value = "ANDROID_SERIAL"
        args = ["a", "b", "c"]
        with patch("subprocess.Popen") as Popen:
            os.name = "posix"
            adb.raw_cmd(*args)
            Popen.assert_called_once_with(
                [" ".join([adb.adb()] + ["-H", "test.com", "-P", "1000"] + args)],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )

    def test_device_serial(self):
        with patch.dict('os.environ', {'ANDROID_SERIAL': "ABCDEF123456"}):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {"ABCDEF123456": "device"}
            self.assertEqual(adb.device_serial(), "ABCDEF123456")
        with patch.dict('os.environ', {'ANDROID_SERIAL': "ABCDEF123456"}):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {"ABCDEF123456": "device", "123456ABCDEF": "device"}
            self.assertEqual(adb.device_serial(), "ABCDEF123456")
        with patch.dict('os.environ', {'ANDROID_SERIAL': "HIJKLMN098765"}):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {"ABCDEF123456": "device", "123456ABCDEF": "device"}
            self.assertEqual(adb.device_serial(), "HIJKLMN098765")
        with patch.dict('os.environ', {}, clear=True):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {"ABCDEF123456": "device", "123456ABCDEF": "device"}
            with self.assertRaises(EnvironmentError):
                adb.device_serial()
        with patch.dict('os.environ', {}, clear=True):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {"ABCDEF123456": "device"}
            self.assertEqual(adb.device_serial(), "ABCDEF123456")

        with self.assertRaises(EnvironmentError):
            adb = Adb()
            adb.devices = MagicMock()
            adb.devices.return_value = {}
            adb.device_serial()

    def test_forward_list(self):
        adb = Adb()
        adb.version = MagicMock()
        adb.version.return_value = ['1.0.31', '1', '0', '31']
        adb.raw_cmd = MagicMock()
        adb.raw_cmd.return_value.communicate.return_value = (b"014E05DE0F02000E    tcp:9008    tcp:9008\r\n489328DKFL7DF    tcp:9008    tcp:9008", b"")
        self.assertEqual(adb.forward_list(), [['014E05DE0F02000E', 'tcp:9008', 'tcp:9008'], ['489328DKFL7DF', 'tcp:9008', 'tcp:9008']])

        adb.version.return_value = ['1.0.29', '1', '0', '29']
        with self.assertRaises(EnvironmentError):
            adb.forward_list()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import re
import os.path
import codecs
from mock import MagicMock, call, patch
from uiautomator import AutomatorDevice, Selector


class TestDevice(unittest.TestCase):

    def setUp(self):
        self.device = AutomatorDevice()
        self.device.server = MagicMock()
        self.device.server.jsonrpc = MagicMock()
        self.device.server.jsonrpc_wrap = MagicMock()

    def test_info(self):
        self.device.server.jsonrpc.deviceInfo = MagicMock()
        self.device.server.jsonrpc.deviceInfo.return_value = {}
        self.assertEqual(self.device.info, {})
        self.device.server.jsonrpc.deviceInfo.assert_called_once_with()

    def test_click(self):
        self.device.server.jsonrpc.click = MagicMock()
        self.device.server.jsonrpc.click.return_value = True
        self.assertEqual(self.device.click(1, 2), True)
        self.device.server.jsonrpc.click.assert_called_once_with(1, 2)

    def test_swipe(self):
        self.device.server.jsonrpc.swipe = MagicMock()
        self.device.server.jsonrpc.swipe.return_value = True
        self.assertEqual(self.device.swipe(1, 2, 3, 4, 100), True)
        self.device.server.jsonrpc.swipe.assert_called_once_with(1, 2, 3, 4, 100)

    def test_long_click(self):
        self.device.server.jsonrpc.swipe = MagicMock()
        self.device.server.jsonrpc.swipe.return_value = True
        x, y = 100, 200
        self.assertEqual(self.device.long_click(x, y), True)
        self.device.server.jsonrpc.swipe.assert_called_once_with(x, y, x+1, y+1, 100)

    def test_drag(self):
        self.device.server.jsonrpc.drag = MagicMock()
        self.device.server.jsonrpc.drag.return_value = True
        self.assertEqual(self.device.drag(1, 2, 3, 4, 100), True)
        self.device.server.jsonrpc.drag.assert_called_once_with(1, 2, 3, 4, 100)

    def test_dump(self):
        self.device.server.jsonrpc.dumpWindowHierarchy = MagicMock()
        with codecs.open(os.path.join(os.path.dirname(__file__), "res", "layout.xml"), "r", encoding="utf8") as f:
            xml = f.read()
            self.device.server.jsonrpc.dumpWindowHierarchy.return_value = xml
            self.assertEqual(self.device.dump("/tmp/test.xml"), xml)
            self.device.server.jsonrpc.dumpWindowHierarchy.assert_called_once_with(True, None)
            self.assertEqual(self.device.dump("/tmp/test.xml", False), xml)

            raw_xml = "".join(re.split(r"\n[ ]*", xml))
            self.device.server.jsonrpc.dumpWindowHierarchy.return_value = raw_xml
            self.assertTrue("\n  " in self.device.dump("/tmp/test.xml"))

    def test_screenshot(self):
        self.device.server.jsonrpc.takeScreenshot = MagicMock()
        self.device.server.jsonrpc.takeScreenshot.return_value = "1.png"
        self.device.server.adb.cmd = cmd = MagicMock()
        self.device.server.screenshot = MagicMock()
        self.device.server.screenshot.return_value = None
        cmd.return_value.returncode = 0
        self.assertEqual(self.device.screenshot("a.png", 1.0, 99), "a.png")
        self.device.server.jsonrpc.takeScreenshot.assert_called_once_with("screenshot.png", 1.0, 99)
        self.assertEqual(cmd.call_args_list, [call("pull", "1.png", "a.png"), call("shell", "rm", "1.png")])

        self.device.server.jsonrpc.takeScreenshot.return_value = None
        self.assertEqual(self.device.screenshot("a.png", 1.0, 100), None)

    def test_freeze_rotation(self):
        self.device.server.jsonrpc.freezeRotation = MagicMock()
        self.device.freeze_rotation(True)
        self.device.freeze_rotation(False)
        self.assertEqual(self.device.server.jsonrpc.freezeRotation.call_args_list, [call(True), call(False)])

    def test_orientation(self):
        self.device.server.jsonrpc.deviceInfo = MagicMock()
        orientation = {
            0: "natural",
            1: "left",
            2: "upsidedown",
            3: "right"
        }
        for i in range(4):
            self.device.server.jsonrpc.deviceInfo.return_value = {"displayRotation": i}
            self.assertEqual(self.device.orientation, orientation[i])
        # set
        orientations = [
            (0, "natural", "n", 0),
            (1, "left", "l", 90),
            (2, "upsidedown", "u", 180),
            (3, "right", "r", 270)
        ]
        for values in orientations:
            for value in values:
                self.device.server.jsonrpc.setOrientation = MagicMock()
                self.device.orientation = value
                self.device.server.jsonrpc.setOrientation.assert_called_once_with(values[1])

        with self.assertRaises(ValueError):
            self.device.orientation = "invalid orientation"

    def test_last_traversed_text(self):
        self.device.server.jsonrpc.getLastTraversedText = MagicMock()
        self.device.server.jsonrpc.getLastTraversedText.return_value = "abcdef"
        self.assertEqual(self.device.last_traversed_text, "abcdef")
        self.device.server.jsonrpc.getLastTraversedText.assert_called_once_with()

    def test_clear_traversed_text(self):
        self.device.server.jsonrpc.clearLastTraversedText = MagicMock()
        self.device.clear_traversed_text()
        self.device.server.jsonrpc.clearLastTraversedText.assert_called_once_with()

    def test_open(self):
        self.device.server.jsonrpc.openNotification = MagicMock()
        self.device.open.notification()
        self.device.server.jsonrpc.openNotification.assert_called_once_with()
        self.device.server.jsonrpc.openQuickSettings = MagicMock()
        self.device.open.quick_settings()
        self.device.server.jsonrpc.openQuickSettings.assert_called_once_with()

    def test_watchers(self):
        names = ["a", "b", "c"]
        self.device.server.jsonrpc.getWatchers = MagicMock()
        self.device.server.jsonrpc.getWatchers.return_value = names
        self.assertEqual(self.device.watchers, names)
        self.device.server.jsonrpc.getWatchers.assert_called_once_with()

        self.device.server.jsonrpc.hasAnyWatcherTriggered = MagicMock()
        self.device.server.jsonrpc.hasAnyWatcherTriggered.return_value = True
        self.assertEqual(self.device.watchers.triggered, True)
        self.device.server.jsonrpc.hasAnyWatcherTriggered.assert_called_once_with()

        self.device.server.jsonrpc.removeWatcher = MagicMock()
        self.device.watchers.remove("a")
        self.device.server.jsonrpc.removeWatcher.assert_called_once_with("a")
        self.device.server.jsonrpc.removeWatcher = MagicMock()
        self.device.watchers.remove()
        self.assertEqual(self.device.server.jsonrpc.removeWatcher.call_args_list, [call(name) for name in names])

        self.device.server.jsonrpc.resetWatcherTriggers = MagicMock()
        self.device.watchers.reset()
        self.device.server.jsonrpc.resetWatcherTriggers.assert_called_once_with()

        self.device.server.jsonrpc.runWatchers = MagicMock()
        self.device.watchers.run()
        self.device.server.jsonrpc.runWatchers.assert_called_once_with()

    def test_watcher(self):
        self.device.server.jsonrpc.hasWatcherTriggered = MagicMock()
        self.device.server.jsonrpc.hasWatcherTriggered.return_value = False
        self.assertFalse(self.device.watcher("name").triggered)
        self.device.server.jsonrpc.hasWatcherTriggered.assert_called_once_with("name")

        self.device.server.jsonrpc.removeWatcher = MagicMock()
        self.device.watcher("a").remove()
        self.device.server.jsonrpc.removeWatcher.assert_called_once_with("a")

        self.device.server.jsonrpc.registerClickUiObjectWatcher = MagicMock()
        condition1 = {"text": "my text", "className": "android"}
        condition2 = {"description": "my desc", "clickable": True}
        target = {"className": "android.widget.Button", "text": "OK"}
        self.device.watcher("watcher").when(**condition1).when(**condition2).click(**target)
        self.device.server.jsonrpc.registerClickUiObjectWatcher.assert_called_once_with(
            "watcher",
            [Selector(**condition1), Selector(**condition2)],
            Selector(**target)
        )

        self.device.server.jsonrpc.registerPressKeyskWatcher = MagicMock()
        self.device.watcher("watcher2").when(**condition1).when(**condition2).press.back.home.power("menu")
        self.device.server.jsonrpc.registerPressKeyskWatcher.assert_called_once_with(
            "watcher2", [Selector(**condition1), Selector(**condition2)], ("back", "home", "power", "menu"))

    def test_press(self):
        key = ["home", "back", "left", "right", "up", "down", "center",
               "menu", "search", "enter", "delete", "del", "recent",
               "volume_up", "volume_down", "volume_mute", "camera", "power"]
        self.device.server.jsonrpc.pressKey = MagicMock()
        self.device.server.jsonrpc.pressKey.return_value = True
        self.assertTrue(self.device.press.home())
        self.device.server.jsonrpc.pressKey.return_value = False
        self.assertFalse(self.device.press.back())
        self.device.server.jsonrpc.pressKey.return_value = False
        for k in key:
            self.assertFalse(self.device.press(k))
        self.assertEqual(self.device.server.jsonrpc.pressKey.call_args_list, [call("home"), call("back")] + [call(k) for k in key])

        self.device.server.jsonrpc.pressKeyCode.return_value = True
        self.assertTrue(self.device.press(1))
        self.assertTrue(self.device.press(1, 2))
        self.assertEqual(self.device.server.jsonrpc.pressKeyCode.call_args_list, [call(1), call(1, 2)])

    def test_wakeup(self):
        self.device.server.jsonrpc.wakeUp = MagicMock()
        self.device.wakeup()
        self.device.server.jsonrpc.wakeUp.assert_called_once_with()

        self.device.server.jsonrpc.wakeUp = MagicMock()
        self.device.screen.on()
        self.device.server.jsonrpc.wakeUp.assert_called_once_with()

        self.device.server.jsonrpc.wakeUp = MagicMock()
        self.device.screen("on")
        self.device.server.jsonrpc.wakeUp.assert_called_once_with()

    def test_screen_status(self):
        self.device.server.jsonrpc.deviceInfo = MagicMock()
        self.device.server.jsonrpc.deviceInfo.return_value = {"screenOn": True}
        self.assertTrue(self.device.screen == "on")
        self.assertTrue(self.device.screen != "off")

        self.device.server.jsonrpc.deviceInfo.return_value = {"screenOn": False}
        self.assertTrue(self.device.screen == "off")
        self.assertTrue(self.device.screen != "on")

    def test_sleep(self):
        self.device.server.jsonrpc.sleep = MagicMock()
        self.device.sleep()
        self.device.server.jsonrpc.sleep.assert_called_once_with()

        self.device.server.jsonrpc.sleep = MagicMock()
        self.device.screen.off()
        self.device.server.jsonrpc.sleep.assert_called_once_with()

        self.device.server.jsonrpc.sleep = MagicMock()
        self.device.screen("off")
        self.device.server.jsonrpc.sleep.assert_called_once_with()

    def test_wait_idle(self):
        self.device.server.jsonrpc_wrap.return_value.waitForIdle = MagicMock()
        self.device.server.jsonrpc_wrap.return_value.waitForIdle.return_value = True
        self.assertTrue(self.device.wait.idle(timeout=10))
        self.device.server.jsonrpc_wrap.return_value.waitForIdle.assert_called_once_with(10)

        self.device.server.jsonrpc_wrap.return_value.waitForIdle = MagicMock()
        self.device.server.jsonrpc_wrap.return_value.waitForIdle.return_value = False
        self.assertFalse(self.device.wait("idle", timeout=10))
        self.device.server.jsonrpc_wrap.return_value.waitForIdle.assert_called_once_with(10)

    def test_wait_update(self):
        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate = MagicMock()
        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate.return_value = True
        self.assertTrue(self.device.wait.update(timeout=10, package_name="android"))
        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate.assert_called_once_with("android", 10)

        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate = MagicMock()
        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate.return_value = False
        self.assertFalse(self.device.wait("update", timeout=100, package_name="android"))
        self.device.server.jsonrpc_wrap.return_value.waitForWindowUpdate.assert_called_once_with("android", 100)

    def test_get_info_attr(self):
        info = {"test_a": 1, "test_b": "string", "displayWidth": 720, "displayHeight": 1024}
        self.device.server.jsonrpc.deviceInfo = MagicMock()
        self.device.server.jsonrpc.deviceInfo.return_value = info
        for k in info:
            self.assertEqual(getattr(self.device, k), info[k])
        self.assertEqual(self.device.width, info["displayWidth"])
        self.assertEqual(self.device.height, info["displayHeight"])
        with self.assertRaises(AttributeError):
            self.device.not_exists

    def test_device_obj(self):
        with patch("uiautomator.AutomatorDeviceObject") as AutomatorDeviceObject:
            kwargs = {"text": "abc", "description": "description...", "clickable": True}
            self.device(**kwargs)
            AutomatorDeviceObject.assert_called_once_with(self.device, Selector(**kwargs))

        with patch("uiautomator.AutomatorDeviceObject") as AutomatorDeviceObject:
            AutomatorDeviceObject.return_value.exists = True
            self.assertTrue(self.device.exists(clickable=True))
            AutomatorDeviceObject.return_value.exists = False
            self.assertFalse(self.device.exists(text="..."))


class TestDeviceWithSerial(unittest.TestCase):

    def test_serial(self):
        with patch('uiautomator.AutomatorServer') as AutomatorServer:
            AutomatorDevice("abcdefhijklmn")
            AutomatorServer.assert_called_once_with(serial="abcdefhijklmn", local_port=None, adb_server_host=None, adb_server_port=None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import uiautomator
import os
from mock import patch


class TestMisc(unittest.TestCase):

    def test_proxy(self):
        self.assertTrue('no_proxy' in os.environ)
        self.assertTrue('localhost' in os.environ.get('no_proxy', ''))

    def test_load(self):
        try:
            from imp import reload
        except:
            pass
        reload(uiautomator)
        self.assertIsNotNone(uiautomator.device)
        self.assertIsNotNone(uiautomator.rect)
        self.assertIsNotNone(uiautomator.point)

    def test_rect(self):
        import random
        for i in range(10):
            top, left = random.randint(0, 100), random.randint(0, 100)
            bottom, right = random.randint(101, 1024), random.randint(101, 720)
            self.assertEqual(uiautomator.rect(top, left, bottom, right), {"top": top, "left": left, "bottom": bottom, "right": right})

    def test_point(self):
        import random
        for i in range(10):
            x, y = random.randint(0, 1024), random.randint(0, 720)
            self.assertEqual(uiautomator.point(x, y), {"x": x, "y": y})

    def test_next_port(self):
        with patch('socket.socket') as socket:
            socket.return_value.connect_ex.side_effect = [0, 0, 1]
            uiautomator._init_local_port = 9007
            self.assertEqual(uiautomator.next_local_port(), 9010)

        with patch('socket.socket') as socket:
            socket.return_value.connect_ex.return_value = 1
            uiautomator._init_local_port = 32764
            self.assertEqual(uiautomator.next_local_port(), 9008)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


requires = [
    "urllib3>=1.7.1"
]
test_requires = [
    'nose>=1.0',
    'mock>=1.0.1',
    'coverage>=3.6'
]

version = '0.3.0'

setup(
    name='uiautomator',
    version=version,
    description='Python Wrapper for Android UiAutomator test tool',
    long_description='Python wrapper for Android uiautomator tool.',
    author='Xiaocong He',
    author_email='xiaocong@gmail.com,hongbin.bao@gmail.com',
    url='https://github.com/xiaocong/uiautomator',
    download_url='https://github.com/xiaocong/uiautomator/tarball/%s' % version,
    keywords=[
        'testing', 'android', 'uiautomator'
    ],
    install_requires=requires,
    tests_require=test_requires,
    test_suite="nose.collector",
    packages=['uiautomator'],
    package_data={
        'uiautomator': [
            'uiautomator/libs/bundle.jar',
            'uiautomator/libs/uiautomator-stub.jar',
            'uiautomator/libs/app-uiautomator-test.apk',
            'uiautomator/libs/app-uiautomator.apk'
        ]
    },
    include_package_data=True,
    license='MIT',
    platforms='any',
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing'
    )
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from uiautomator import Selector


class TestSelector(unittest.TestCase):

    fields = {
        "text": (0x01, None),  # MASK_TEXT,
        "textContains": (0x02, None),  # MASK_TEXTCONTAINS,
        "textMatches": (0x04, None),  # MASK_TEXTMATCHES,
        "textStartsWith": (0x08, None),  # MASK_TEXTSTARTSWITH,
        "className": (0x10, None),  # MASK_CLASSNAME
        "classNameMatches": (0x20, None),  # MASK_CLASSNAMEMATCHES
        "description": (0x40, None),  # MASK_DESCRIPTION
        "descriptionContains": (0x80, None),  # MASK_DESCRIPTIONCONTAINS
        "descriptionMatches": (0x0100, None),  # MASK_DESCRIPTIONMATCHES
        "descriptionStartsWith": (0x0200, None),  # MASK_DESCRIPTIONSTARTSWITH
        "checkable": (0x0400, False),  # MASK_CHECKABLE
        "checked": (0x0800, False),  # MASK_CHECKED
        "clickable": (0x1000, False),  # MASK_CLICKABLE
        "longClickable": (0x2000, False),  # MASK_LONGCLICKABLE,
        "scrollable": (0x4000, False),  # MASK_SCROLLABLE,
        "enabled": (0x8000, False),  # MASK_ENABLED,
        "focusable": (0x010000, False),  # MASK_FOCUSABLE,
        "focused": (0x020000, False),  # MASK_FOCUSED,
        "selected": (0x040000, False),  # MASK_SELECTED,
        "packageName": (0x080000, None),  # MASK_PACKAGENAME,
        "packageNameMatches": (0x100000, None),  # MASK_PACKAGENAMEMATCHES,
        "resourceId": (0x200000, None),  # MASK_RESOURCEID,
        "resourceIdMatches": (0x400000, None),  # MASK_RESOURCEIDMATCHES,
        "index": (0x800000, 0),  # MASK_INDEX,
        "instance": (0x01000000, 0)  # MASK_INSTANCE,
    }
    mask = "mask"

    def test_init(self):
        sel = Selector()
        self.assertEqual(sel[self.mask], 0)
        self.assertEqual(sel["childOrSibling"], [])
        self.assertEqual(sel["childOrSiblingSelector"], [])

    def test_add(self):
        for k, v in self.fields.items():
            kwargs = {k: v[1]}
            sel = Selector(**kwargs)
            self.assertEqual(sel[self.mask], v[0])

        for k1, v1 in self.fields.items():
            for k2, v2 in self.fields.items():
                if k1 != k2:
                    kwargs = {k1: v1[1], k2: v2[1]}
                    sel = Selector(**kwargs)
                    self.assertEqual(sel[self.mask], v1[0] | v2[0])

    def test_delete(self):
        for k, v in self.fields.items():
            kwargs = {k: v[1]}
            sel = Selector(**kwargs)
            del sel[k]
            self.assertEqual(sel[self.mask], 0)

        for k1, v1 in self.fields.items():
            for k2, v2 in self.fields.items():
                if k1 != k2:
                    kwargs = {k1: v1[1], k2: v2[1]}
                    sel = Selector(**kwargs)
                    del sel[k1]
                    self.assertEqual(sel[self.mask], v2[0])
                    del sel[k2]
                    self.assertEqual(sel[self.mask], 0)

    def test_error(self):
        with self.assertRaises(ReferenceError):
            Selector(text1="")

    def test_child_and_sibling(self):
        sel = Selector()
        sel.child(text="...")
        self.assertEqual(sel["childOrSibling"], ["child"])
        self.assertEqual(sel["childOrSiblingSelector"], [Selector(text="...")])

        sel.sibling(text="---")
        self.assertEqual(sel["childOrSibling"], ["child", "sibling"])
        self.assertEqual(sel["childOrSiblingSelector"], [Selector(text="..."), Selector(text="---")])

    def test_clone(self):
        kwargs = {
            "text": "1234",
            "description": "desc...",
            "clickable": True,
            "focusable": False,
            "packageName": "android"
        }
        sel = Selector(**kwargs)
        sel.child(text="1")
        sel.sibling(text="1")
        sel.child(text="1")

        clone = sel.clone()
        for k in kwargs:
            self.assertEqual(sel[k], clone[k])
        self.assertEqual(sel["childOrSibling"], clone["childOrSibling"])
        self.assertEqual(sel["childOrSiblingSelector"], clone["childOrSiblingSelector"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from mock import MagicMock, patch
from uiautomator import JsonRPCMethod, JsonRPCClient
import os


class TestJsonRPCMethod_id(unittest.TestCase):

    def test_id(self):
        method = JsonRPCMethod("", "method", 30)
        self.assertTrue(isinstance(method.id(), str))
        self.assertTrue(len(method.id()) > 0)
        for i in range(100):
            self.assertNotEqual(method.id(), method.id())


class TestJsonRPCMethod_call(unittest.TestCase):

    def setUp(self):
        self.url = "http://localhost/jsonrpc"
        self.timeout = 20
        self.method_name = "ping"
        self.id = "fGasV62G"
        self.method = JsonRPCMethod(self.url, self.method_name, self.timeout)
        self.method.id = MagicMock()
        self.method.id.return_value = self.id
        try:
            import urllib2
            self.urlopen_patch = patch('urllib2.urlopen')
        except:
            self.urlopen_patch = patch('urllib.request.urlopen')
        finally:
            self.urlopen = self.urlopen_patch.start()

    def tearDown(self):
        self.urlopen_patch.stop()

    def test_normal_call(self):
        return_mock = self.urlopen.return_value
        return_mock.getcode.return_value = 200

        return_mock.read.return_value = b'{"result": "pong", "error": null, "id": "DKNCJDLDJJ"}'
        self.assertEqual("pong", self.method())
        self.method.id.assert_called_once_with()

        return_mock.read.return_value = b'{"result": "pong", "id": "JDLSFJLILJEMNC"}'
        self.assertEqual("pong", self.method())
        self.assertEqual("pong", self.method(1, 2, "str", {"a": 1}, ["1"]))
        self.assertEqual("pong", self.method(a=1, b=2))

    def test_normal_call_error(self):
        return_mock = self.urlopen.return_value

        return_mock.getcode.return_value = 500
        with self.assertRaises(Exception):
            self.method()

        return_mock.getcode.return_value = 200
        return_mock.read.return_value = b'{"result": "pong", "error": {"code": -513937, "message": "error message."}, "id": "fGasV62G"}'
        with self.assertRaises(Exception):
            self.method()
        return_mock.read.assert_called_with()

        return_mock.getcode.return_value = 200
        return_mock.read.return_value = b'{"result": null, "error": null, "id": "fGasV62G"}'
        with self.assertRaises(SyntaxError):
            self.method(1, 2, kwarg1="")


class TestJsonRPCClient(unittest.TestCase):

    def setUp(self):
        self.url = "http://localhost/jsonrpc"
        self.timeout = 20

    def test_jsonrpc(self):
        with patch('uiautomator.JsonRPCMethod') as JsonRPCMethod:
            client = JsonRPCClient(self.url, self.timeout, JsonRPCMethod)
            JsonRPCMethod.return_value = "Ok"
            self.assertEqual(client.ping, "Ok")
            JsonRPCMethod.assert_called_once_with(self.url, "ping", timeout=self.timeout)

            JsonRPCMethod.return_value = {"width": 10, "height": 20}
            self.assertEqual(client.info, {"width": 10, "height": 20})
            JsonRPCMethod.assert_called_with(self.url, "info", timeout=self.timeout)


class TestJsonRPCMethod_call_on_windows(unittest.TestCase):

    def setUp(self):
        self.os_name = os.name
        os.name = "nt"
        self.url = "http://localhost/jsonrpc"
        self.timeout = 20
        self.method_name = "ping"
        self.id = "fGasV62G"
        self.method = JsonRPCMethod(self.url, self.method_name, self.timeout)
        self.method.pool = MagicMock()
        self.method.id = MagicMock()
        self.method.id.return_value = self.id

    def tearDown(self):
        os.name = self.os_name

    def test_normal_call(self):
        urlopen = self.method.pool.urlopen
        urlopen.return_value.status = 200

        urlopen.return_value.data = b'{"result": "pong", "error": null, "id": "DKNCJDLDJJ"}'
        self.assertEqual("pong", self.method())
        self.method.id.assert_called_once_with()

        urlopen.return_value.data = b'{"result": "pong", "id": "JDLSFJLILJEMNC"}'
        self.assertEqual("pong", self.method())
        self.assertEqual("pong", self.method(1, 2, "str", {"a": 1}, ["1"]))
        self.assertEqual("pong", self.method(a=1, b=2))

    def test_normal_call_error(self):
        urlopen = self.method.pool.urlopen
        urlopen.return_value.status = 500

        with self.assertRaises(Exception):
            self.method()
